import sys

import unittest, unittest.mock
import struct
import binascii

# Check for earlier possible mocking of the usb.core and usb.util modules and delete those
if "usb.core" in sys.modules.keys():
    del sys.modules['usb.core'], sys.modules['usb.core.find'], sys.modules['usb.util']


class EndpointMock:
    """Mock values from usb.util module. See inline comments for attribute name."""
    bmAttributes: int = 0x03  # ENDPOINT_TYPE_INTR
    bEndpointAddress_in = 0x80  # ENDPOINT_IN
    bEndpointAddress_out = 0x00  # ENDPOINT_OUT

    def __init__(self, endpoint=True):
        if endpoint:
            self.bEndpointAddress = self.bEndpointAddress_in
        else:
            self.bEndpointAddress = self.bEndpointAddress_out


class InterfaceClassMock:
    """Mock endpoints of an active device"""
    bInterfaceClass: int = 3
    index = 0

    def __init__(self, endpoint_in, endpoint_out):
        self._endpoint_in = endpoint_in
        self._endpoint_out = endpoint_out

    def endpoints(self):
        mock_in = EndpointMock(self._endpoint_in)
        mock_out = EndpointMock(not self._endpoint_out)

        return [mock_in, mock_out]


class InterfaceMock:
    """Mock active configuration interfaces of a device"""
    def __init__(self, no_active_config=False, endpoint_in=True, endpoint_out=True):
        self._no_active_config = no_active_config
        self._endpoint_in = endpoint_in
        self._endpoint_out = endpoint_out

    def interfaces(self):
        if self._no_active_config:
            return []

        return [InterfaceClassMock(self._endpoint_in, self._endpoint_out)]


class DevMock:
    """Mock active configuration call."""
    def __init__(self, no_active_config=False, endpoint_in=True, endpoint_out=True):
        """
        Parameters:
            no_active_config: get_active_configuration will return an empty list if True.
            endpoint_in: By setting this as False the returned interface has wrong in endpoint value.
            endpoint_out: By setting this as False the returned interface has wrong out endpoint value.
        """
        self._no_active_config = no_active_config
        self._endpoint_in = endpoint_in
        self._endpoint_out = endpoint_out

    def get_active_configuration(self):
        """Return an active configuration."""
        return InterfaceMock(self._no_active_config, self._endpoint_in, self._endpoint_out)

    def is_kernel_driver_active(self, index):
        """Mock detach of kernel driver"""
        if index == 0:
            return False  # No detaching
        else:
            return True  # Detaching

    def ctrl_transfer(self, *args):
        return

    def read(self, *args):
        raise _usb_core_mock.USBError


# mock the later on imported usb.core and usb.util
_usb_core_find_mock = unittest.mock.Mock(return_value=DevMock())
_usb_core_mock = unittest.mock.Mock()
_usb_util_mock = unittest.mock.Mock()
sys.modules['usb.core'] = _usb_core_mock
sys.modules['usb.core.find'] = _usb_core_find_mock
sys.modules['usb.util'] = _usb_util_mock
sys.modules['usb.util.claim_interface'] = unittest.mock.Mock()

from qmi.instruments.thorlabs import Thorlabs_Tsp01b
import qmi.core.exceptions
from qmi.utils.context_managers import open_close

ENDPOINT_TYPE_INTR: int = 0x03
ENDPOINT_IN: int = 0x80
ENDPOINT_OUT: int = 0x00


# Mock functions
def endpoint_type(attribute):
    """Mock usb.util.endpoint_type call"""
    return ENDPOINT_TYPE_INTR


def endpoint_direction(direction: int):
    """Mock usb.util.endpoint_direction call"""
    if direction == EndpointMock.bEndpointAddress_in:
        return EndpointMock.bEndpointAddress_in

    elif direction == EndpointMock.bEndpointAddress_out:
        return EndpointMock.bEndpointAddress_out


class TestThorlabsTsp01b(unittest.TestCase):
    def setUp(self):
        _usb_core_mock.find.detach_kernel_driver = unittest.mock.Mock()
        self.patch_error = unittest.mock.patch("usb.core.USBError", BaseException)
        self.patch_claim = unittest.mock.patch("usb.util.claim_interface")
        self.patch_dispose = unittest.mock.patch("usb.util.dispose_resources")
        self.patch_error.start()
        self.patch_claim.start()
        self.patch_dispose.start()
        _usb_util_mock.ENDPOINT_TYPE_INTR = ENDPOINT_TYPE_INTR
        _usb_util_mock.ENDPOINT_IN = ENDPOINT_IN
        _usb_util_mock.ENDPOINT_OUT = ENDPOINT_OUT
        _usb_util_mock.endpoint_type = endpoint_type
        _usb_util_mock.endpoint_direction = endpoint_direction

        qmi.start("TestTsp01bContext")
        self._serial = "123456"
        self.instr: Thorlabs_Tsp01b = qmi.make_instrument("instr", Thorlabs_Tsp01b, self._serial)

    def tearDown(self):
        qmi.stop()
        self.patch_dispose.stop()
        self.patch_claim.stop()
        self.patch_error.stop()

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_open_close(self, usbmock, patch_find):
        """Test opening and closing the instrument"""
        self.instr.open()
        self.instr.close()

        patch_find.assert_called_with(
            find_all=False, idVendor=0x1313, idProduct=0x80fa, serial_number=self._serial)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=None)
    def test_open_close_except_with_no_serial_number(self, usbmock, patch_find):
        """Test opening the instrument with no serial number excepts"""
        expected = f"Instrument with serial number {self._serial} not found (check device permission)"
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.open()

        self.assertEqual(str(exc.exception), expected)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock(True))
    def test_open_close_except_with_no_HID_interface(self, usbmock, patch_find):
        """Test opening the instrument with no active config excepts"""
        expected = "Instrument does not have a HID interface"
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.open()

        self.assertEqual(str(exc.exception), expected)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock(False, False, True))
    def test_open_close_except_with_no_IN_endpoint(self, usbmock, patch_find):
        """Test opening the instrument excepts if IN endpoint is not correctly defined."""
        expected = "Instrument does not have Interrupt IN endpoint"
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.open()

        self.assertEqual(str(exc.exception), expected)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock(False, True, False))
    def test_open_close_except_with_no_OUT_endpoint(self, usbmock, patch_find):
        """Test opening the instrument excepts if OUT endpoint is not correctly defined."""
        expected = "Instrument does not have Interrupt OUT endpoint"
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.open()

        self.assertEqual(str(exc.exception), expected)


class TestThorlabsTsp01bMethods(unittest.TestCase):

    def _create_byte_string(self, code, data, cmd=None):
        # Packet format (32 bytes length):
        #   byte 0: fixed byte 0xF0
        #   byte 1: data length
        #   byte 2: fixed byte 0x00
        #   byte 3: fixed byte 0x01
        #   byte 4: reply/command code
        #   byte 5: CRC-8 over first 5 bytes
        #   byte 6: fixed byte 0xF1
        #   starting at byte 7: variable length data (may be empty)
        #   following data: 4 bytes CRC-32 over data (excluding the 0xF1 marker)
        #   padding with 0x00 until total length 32 bytes

        # bytes 0-4
        if cmd:
            # Data length is 0
            byte_string = bytes([0xf0, 0x00, 0x00, 0x01, code])
            # In the write call the command byte gets replaced with the ASCII version
            crc8_byte = cmd

        else:
            byte_string = bytes([0xf0, len(data), 0x00, 0x01, code])
            # I use the same static method for the CRC-8-AUTOSAR calculation from the code
            crc8_byte = bytes([Thorlabs_Tsp01b._crc8(byte_string)])

        # bytes 5-6
        byte_string += crc8_byte + bytes([0xf1])
        if cmd:
            # We do not CRC-check commands. So, just put zeroes
            data_crc = b"\x00" * 4

        else:
            # Calculate CRC-32 over data.
            data_crc = struct.pack("<I", binascii.crc32(data))

        # Add zero-padding until 32 bytes total length.
        padding = bytes((21 - len(data)) * [0x00])
        # Format final packet.
        byte_string = byte_string + data + data_crc + padding
        return byte_string

    def setUp(self):
        DevMock.write = unittest.mock.Mock()
        self.patch_error = unittest.mock.patch("usb.core.USBError", BaseException)
        self.patch_claim = unittest.mock.patch("usb.util.claim_interface")
        self.patch_dispose = unittest.mock.patch("usb.util.dispose_resources")
        self.patch_error.start()
        self.patch_claim.start()
        self.patch_dispose.start()
        _usb_util_mock.ENDPOINT_TYPE_INTR = ENDPOINT_TYPE_INTR
        _usb_util_mock.ENDPOINT_IN = ENDPOINT_IN
        _usb_util_mock.ENDPOINT_OUT = ENDPOINT_OUT
        _usb_util_mock.endpoint_type = endpoint_type
        _usb_util_mock.endpoint_direction = endpoint_direction
        qmi.start("TestTsp01bContext")
        self._serial = "123456"
        self.instr: Thorlabs_Tsp01b = qmi.make_instrument("instr", Thorlabs_Tsp01b, self._serial)

    def tearDown(self):
        self.patch_dispose.stop()
        self.patch_claim.stop()
        self.patch_error.stop()
        qmi.stop()

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_reset(self, patch_find):
        """Test the reset method."""
        command_string = self._create_byte_string(0x01, bytes([0x00]), bytes([189]))
        reply_string = self._create_byte_string(0x04, bytes([0x00]))
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string]
        with open_close(self.instr):
            self.instr.reset()

        DevMock.write.assert_called_with(0, command_string, 2000)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_unexpected_reply_to_reset(self, patch_find):
        """Test the reset method raises exception."""
        expected = "Unexpected reply to reset command (01, 00)"
        # Now we return wrong reply parameter
        reply_string = self._create_byte_string(0x01, bytes([0x00]))
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string]
        with open_close(self.instr), self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.reset()

        self.assertEqual(str(exc.exception), expected)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_get_idn(self, patch_find):
        """Test the get_idn method."""
        expected_idn = []
        command_strings, reply_strings = [], []
        for i in [3, 2, 4, 1]:
            expected_idn.append(i)
            command_strings.append(
                self._create_byte_string(0x01, bytes([0x00]), bytes([189]))
            )
            reply_strings.append(
                self._create_byte_string(0x04, bytes([0x00 + i, 0x00 + i]))
            )
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException] + reply_strings
        with open_close(self.instr):
            idn = self.instr.get_idn()

        for i in range(4):
            DevMock.write.assert_called_with(0, command_strings[i], 2000)

        self.assertEqual(ord(idn.vendor), expected_idn[0])
        self.assertEqual(ord(idn.model), expected_idn[1])
        self.assertEqual(ord(idn.serial), expected_idn[2])
        self.assertEqual(ord(idn.version), expected_idn[3])

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_unexpected_reply_to_device_info_query(self, patch_find):
        """Test the get_idn raises an exception."""
        expected = "Unexpected reply to device info query (01, 00)"
        # Now we return wrong reply parameter
        reply_string = self._create_byte_string(0x01, bytes([0x00]))
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string]
        with open_close(self.instr), self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.get_idn()

        self.assertEqual(str(exc.exception), expected)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_get_internal_temperature(self, patch_find):
        """Test get_internal_temperature method returns temperature value."""
        expected_t = -273.0  # Quite cold!
        command_string = self._create_byte_string(0x01, bytes([0x00]), bytes([189]))
        # Now length of reply data should be 5 and start with channel number, see driver.
        reply_string = self._create_byte_string(0x07, b"\x00\x00\x80\x88\xc3")
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string]
        with open_close(self.instr):
            temp_int = self.instr.get_internal_temperature()

        DevMock.write.assert_called_with(0, command_string, 2000)
        self.assertEqual(temp_int, expected_t)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_sensor_not_connected(self, patch_find):
        """Test get_internal_temperature method raises exception."""
        expected = "Sensor not connected"
        # Reply parameter 0x13 indicates sensor is not connected, with specific byte string
        reply_string = self._create_byte_string(0x13, b"\x03\x02\x24\x00")
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string]
        with open_close(self.instr), self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.get_internal_temperature()

        self.assertEqual(str(exc.exception), expected)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_unexpected_reply_to_measurement_query(self, patch_find):
        """Test get_internal_temperature method raises exception."""
        expected = "Unexpected reply to measurement query (07, 00 00 00 00)"
        # Reply data misses one byte, the channel number
        reply_string = self._create_byte_string(0x07, b"\x00\x00\x00\x00")
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string]
        with open_close(self.instr), self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.get_internal_temperature()

        self.assertEqual(str(exc.exception), expected)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_get_external_temperature(self, patch_find):
        """Test get_internal_temperature method returns temperature value."""
        expected_t = 273.0  # Quite warm!
        channels = [1, 2]
        command_string = self._create_byte_string(0x01, bytes([0x00]), bytes([189]))
        # Now length of reply data should be 5 and start with channel number, see driver.
        reply_string = self._create_byte_string(0x07, b"\x02\x00\x80\x88\x43")
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string, reply_string]
        with open_close(self.instr):
            for channel in channels:
                temp_int = self.instr.get_external_temperature(channel)
                self.assertEqual(temp_int, expected_t)

        DevMock.write.assert_called_with(0, command_string, 2000)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_get_external_temperature_wrong_channel_number(self, patch_find):
        """See that the call excepts on wrong channel number"""
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException]
        with open_close(self.instr), self.assertRaises(ValueError):
            self.instr.get_external_temperature(0)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_get_humidity(self, patch_find):
        """Test get_internal_temperature method returns temperature value."""
        expected_h2o = 0.0  # Dry
        command_string = self._create_byte_string(0x01, bytes([0x00]), bytes([189]))
        # Now length of reply data should be 5 and start with channel number, see driver.
        reply_string = self._create_byte_string(0x07, b"\x01\x00\x00\x00\x00")
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string]
        with open_close(self.instr):
            humidity = self.instr.get_humidity()

        DevMock.write.assert_called_with(0, command_string, 2000)
        self.assertEqual(humidity, expected_h2o)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_packet_too_short_exception(self, patch_find):
        """See that an exception is raised if the packet is too short."""
        expected = "Received short packet from instrument"
        reply_string = self._create_byte_string(0x00, bytes([0x00]), bytes([189]))
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string[:10]]
        with open_close(self.instr), self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.reset()

        self.assertEqual(str(exc.exception)[:len(expected)], expected)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_packet_has_wrong_fixed_values(self, patch_find):
        """See that an exception is raised if the packet has a wrong fixed value."""
        expected = "Received invalid packet from instrument"
        reply_string = self._create_byte_string(0x00, bytes([0x00]), bytes([189]))
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string[1:]]
        with open_close(self.instr), self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.reset()

        self.assertEqual(str(exc.exception)[:len(expected)], expected)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_packet_except_on_bad_crc_check_value(self, patch_find):
        """See that an exception is raised if the CRC8 check value does not match."""
        expected = "Received bad header CRC from instrument"
        reply_string = self._create_byte_string(0x00, bytes([0x00]), bytes([189]))
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string]
        with open_close(self.instr), self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.reset()

        self.assertEqual(str(exc.exception)[:len(expected)], expected)

    @unittest.mock.Mock("qmi.instruments.thorlabs.tsp01b.usb")
    @unittest.mock.patch("qmi.instruments.thorlabs.tsp01b.usb.core.find", return_value=DevMock())
    def test_inconsistent_data_length_raises_exception(self, patch_find):
        """See that an exception is raised if data length is inconsistent with the packet length."""
        expected = "Received invalid packet from instrument"
        reply_string = self._create_byte_string(0x04, bytes([0x00]))
        DevMock.read = unittest.mock.Mock()
        DevMock.read.side_effect = [BaseException, reply_string[:11]]
        with open_close(self.instr), self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.reset()

        self.assertEqual(str(exc.exception)[:len(expected)], expected)


if __name__ == '__main__':
    unittest.main()



"""QMI driver for the NKT Photonics Instruments "Adjustik" laser driver.

The communication protocol to the laser driver is documented in the "Software Development Kit for NKT Photonics
Instruments - Instruction Manual" document, v2.1.3, Oct 2018. [Reference 1].

Chapter 2 of this document describes the NKT Photonics Interbus protocol (physical, framing, CRC, etc.)

module type      standard address         section     description

20h              0Fh                      6.2         Koheras AdjustiK/BoostiK System (K81-1 to K83-1)
34h              80h                      6.3         Koheras ADJUSTIK/ACOUSTIK System (K822/K852)
21h              0Ah                      6.4         Koheras BasiK Module (K80-1)
36h              01h (default; settable)  6.5         Koheras BASIK MIKRO Module (K0x2)
33h              01h (default; settable)  6.6         Koheras BASIK MIKRO Module (K1x2)
70h              02h                      6.7         BoostiK OEM Amplifier (N83)
60h              15h                      6.7         BoostiK OEM Amplifier (N83)

Chapter 6 of this document describes the Register Files for different devices.
"""

import logging
import struct
from typing import Tuple

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport
from qmi.instruments.nkt_photonics.nkt_photonics_interbus_protocol import NKTPhotonicsInterbusProtocol

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class KoherasAdjustikLaser(QMI_Instrument):

    # Unfortunately the laser does not always respond.
    # Most requests get a response within 0.1 seconds.
    # If the laser responds at all, it almost always responds within 0.8 seconds.
    # When this timeout is exceeded, the request will be retried.
    RESPONSE_TIMEOUT = 1.0

    BASIK_ADDRESS = 0x01
    ADJUSTIK_ADDRESS = 0x80

    def __init__(self, context: QMI_Context, name: str, transport: str):
        super().__init__(context, name)
        self._transport = create_transport(transport,
                                           default_attributes={
                                               "baudrate": 115200,
                                               "bytesize": 8,
                                               "parity": 'N',
                                               "stopbits": 1.0,
                                               "rtscts": True})
        self._interbus = NKTPhotonicsInterbusProtocol(self._transport, self.RESPONSE_TIMEOUT)

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        self._transport.discard_read()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to %s", self._name)
        super().close()
        self._transport.close()

    # BASIK module interface, as described in Section 6.6 of the documentation.
    # The BASIK module refers to a specific laser controller.
    # 6.6.1 : BASIK module --- general settings.

    @rpc_method
    def get_basik_module_type(self) -> int:
        """Return BASIK module type. Should be 0x33."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x61)
        unpacked_response = struct.unpack("<B", response)[0]
        return unpacked_response

    @rpc_method
    def get_basik_emission(self) -> int:
        """Emission status."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x30)
        unpacked_response = struct.unpack("<B", response)[0]
        return unpacked_response

    def _set_basik_emission(self, status: bool) -> None:
        """
        Set emission status.

        Parameters:
            status: the status of the laser to be set. True for on and False for off.
        """
        value = struct.pack("<B", int(status))
        self._interbus.set_register(self.BASIK_ADDRESS, 0x30, value)

    @rpc_method
    def enable_basik_emission(self) -> None:
        """
        Enable basik emission.
        """
        self._set_basik_emission(True)

    @rpc_method
    def disable_basik_emission(self) -> None:
        """
        Disable basik emission.
        """
        self._set_basik_emission(False)

    @rpc_method
    def get_basik_setup_bits(self) -> int:
        """Return the setup bits.

        bit 0: -
        bit 1: wide wavelength modulation range
        bit 2: enable external wavelength modulation
        bit 3: wavelength modulation DC coupled
        bit 4: enable internal wavelength modulation
        bit 5: enable modulation output
        bit 6: -
        bit 7: -
        bit 8: pump operation constant current
        bit 9: external amplitude modulation source
        """
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x31)
        unpacked_response = struct.unpack("<H", response)[0]
        return unpacked_response

    @rpc_method
    def set_basik_setup_bits(self, bits: int) -> None:
        """Set the setup bits."""
        value = struct.pack("<H", bits)
        self._interbus.set_register(self.BASIK_ADDRESS, 0x31, value)

    @rpc_method
    def get_basik_output_power_setpoint_mW(self) -> float:
        """Return the output power setpoint, in [mW]."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x22)
        unpacked_response = struct.unpack("<H", response)[0] / 100.0
        return unpacked_response

    @rpc_method
    def set_basik_output_power_setpoint_mW(self, power: float) -> None:
        """Set the the output power setpoint, in [mW]."""
        value = struct.pack("<H", int(round(power * 100.0)))
        self._interbus.set_register(self.BASIK_ADDRESS, 0x22, value)

    @rpc_method
    def get_basik_output_power_setpoint_dBm(self) -> float:
        """Return the output power setpoint, in [dBm].

        Should be equal to 10.0 * log10(output_power_setpoint_mW).
        """
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0xa0)
        unpacked_response = struct.unpack("<h", response)[0] / 100.0
        return unpacked_response

    @rpc_method
    def set_basik_output_power_setpoint_dBm(self, power: float) -> None:
        """Set the output power setpoint, in [dBm].

        Should be equal to 10.0 * log10(output_power_setpoint_mW).
        """
        value = struct.pack("<h", int(round(power * 100.0)))
        self._interbus.set_register(self.BASIK_ADDRESS, 0xa0, value)

    @rpc_method
    def get_basik_wavelength_offset_setpoint(self) -> float:
        """Return the wavelength offset setpoint, in [pm]."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x2a)
        unpacked_response = struct.unpack("<h", response)[0] / 10.0
        return unpacked_response

    @rpc_method
    def set_basik_wavelength_offset_setpoint(self, setpoint: float) -> None:
        """Set the wavelength offset setpoint, in [pm]."""
        setpoint_int = int(round(setpoint * 10.0))
        value = struct.pack("<h", setpoint_int)

        self._interbus.set_register(self.BASIK_ADDRESS, 0x2a, value)

    @rpc_method
    def get_basik_user_area(self) -> bytes:
        """Return the 240-byte user area contents."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x8d)
        return response  # Return raw 'bytes' instance. No unpacking needed.

    # 6.6.2 : BASIK module --- readouts.

    @rpc_method
    def get_basik_status_bits(self) -> int:
        """Return BASIC module type."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x66)
        unpacked_response = struct.unpack("<H", response)[0]
        return unpacked_response

    @rpc_method
    def get_basik_output_power_mW(self) -> float:
        """Return the output power readout, in [mW]."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x17)
        unpacked_response = struct.unpack("<H", response)[0] / 100.0
        return unpacked_response

    @rpc_method
    def get_basik_output_power_dBm(self) -> float:
        """Return the output power readout, in [dBm].

        Should be equal to 10.0 * log10(output_power_mW).
        """
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x90)
        unpacked_response = struct.unpack("<h", response)[0] / 100.0
        return unpacked_response

    @rpc_method
    def get_basik_standard_wavelength(self) -> float:
        """Return the standard wavelength, in [pm]."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x32)
        unpacked_response = struct.unpack("<L", response)[0] / 10.0
        return unpacked_response

    @rpc_method
    def get_basik_wavelength_offset(self) -> float:
        """Return the wavelength offset, in [pm]."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x72)
        unpacked_response = struct.unpack("<l", response)[0] / 10.0
        return unpacked_response

    @rpc_method
    def get_basik_module_temperature(self) -> float:
        """Return the module temperature, in [°C]."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x1c)
        unpacked_response = struct.unpack("<H", response)[0] / 10.0
        return unpacked_response

    @rpc_method
    def get_basik_module_supply_voltage(self) -> float:
        """Return the module supply voltage, in [V]."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x1e)
        unpacked_response = struct.unpack("<H", response)[0] / 1000.0
        return unpacked_response

    @rpc_method
    def get_basik_module_wavelength_modulation_enabled(self) -> int:
        """Undocumented: wavelength modulation on?"""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0xb5)
        unpacked_response = struct.unpack("<B", response)[0]
        return unpacked_response

    # 6.6.3 : BASIK module --- modulation.

    @rpc_method
    def get_basik_wavelength_modulation_frequency(self) -> Tuple[float, float]:
        """Return the two wavelength modulation frequencies, in ([Hz], [Hz])."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0xb8)
        unpacked_response = struct.unpack("<ff", response)
        return unpacked_response  # type: ignore

    @rpc_method
    def set_basik_wavelength_modulation_frequency(self, freq0: float, freq1: float) -> None:
        """Set the wavelength modulation frequency, in [Hz]."""
        value = struct.pack("<ff", freq0, freq1)
        self._interbus.set_register(self.BASIK_ADDRESS, 0xb8, value)

    @rpc_method
    def get_basik_wavelength_modulation_level(self) -> float:
        """Return the wavelength modulation level, in [%]."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x2b)
        unpacked_response = struct.unpack("<H", response)[0] / 10.0
        return unpacked_response

    @rpc_method
    def set_basik_wavelength_modulation_level(self, level: float) -> None:
        """Set the wavelength modulation level, in [%]."""
        level_int = int(round(level * 10.0))
        value = struct.pack("<H", level_int)

        self._interbus.set_register(self.BASIK_ADDRESS, 0x2b, value)

    @rpc_method
    def get_basik_wavelength_modulation_offset(self) -> float:
        """Return the wavelength modulation offset, in [%]."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x2f)
        unpacked_response = struct.unpack("<h", response)[0] / 10.0
        return unpacked_response

    @rpc_method
    def set_basik_wavelength_modulation_offset(self, offset: float) -> None:
        """Set the wavelength modulation offset, in [%]."""
        offset_int = int(round(offset * 10.0))
        value = struct.pack("<h", offset_int)
        self._interbus.set_register(self.BASIK_ADDRESS, 0x2f, value)

    @rpc_method
    def get_basik_amplitude_modulation_frequency(self) -> Tuple[float, float]:
        """Return the two amplitude modulation frequencies, in ([Hz], [Hz])."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0xba)
        unpacked_response = struct.unpack("<ff", response)
        return unpacked_response  # type:ignore

    @rpc_method
    def get_basik_amplitude_modulation_depth(self) -> float:
        """Return the amplitude modulation depth, in [%]."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0x2c)
        unpacked_response = struct.unpack("<H", response)[0] / 10.0
        return unpacked_response

    @rpc_method
    def get_basik_modulation_setup_bits(self) -> int:
        """Return the modulation setup bits."""
        response = self._interbus.get_register(self.BASIK_ADDRESS, 0xb7)
        unpacked_response = struct.unpack("<H", response)[0]
        return unpacked_response

    @rpc_method
    def set_basik_modulation_setup_bits(self, bits: int) -> None:
        """Set the modulation setup bits."""
        value = struct.pack("<H", bits)
        self._interbus.set_register(self.BASIK_ADDRESS, 0xb7, value)

    # ADJUSTIK module interface, as described in Section 6.3 of the documentation.
    # The ADJUSTIK module controls the frame (e.g., multiple BASIK units).

    # 6.3.1 : ADJUSTIK module --- general settings.

    @rpc_method
    def get_adjustik_module_type(self) -> int:
        """Return ADJUSTIK module type. Should be 0x34."""
        response = self._interbus.get_register(self.ADJUSTIK_ADDRESS, 0x61)
        unpacked_response = struct.unpack("<B", response)[0]
        return unpacked_response



"""Extension of the QMI_UsbTmcTransport class utilizing usb.core and qmi.core.usbtmc packages."""

from typing import Optional

import usb.core

from qmi.core import usbtmc as usbtmc
from qmi.core.exceptions import QMI_TimeoutException
from qmi.core.transport import QMI_UsbTmcTransport


class QMI_PyUsbTmcTransport(QMI_UsbTmcTransport):

    def __init__(self, vendorid: int, productid: int, serialnr: str) -> None:
        super().__init__(vendorid, productid, serialnr)
        self._device: Optional[usbtmc.Instrument] = None

    def _open_transport(self) -> None:
        self._device = usbtmc.Instrument(self.vendorid, self.productid, self.serialnr)
        self._device.open()

    def close(self) -> None:
        super().close()
        if self._device is not None:
            self._device.close()

    @property
    def _safe_device(self) -> usbtmc.Instrument:
        """ The _safe_device property should be used inside the QMI_Transport code if-and-only-if we are 100% sure that
        the _device attribute is not None.

        This aids in static typechecking, since whereas the type of _device is Optional[T], the result of this method is
        guaranteed to be of type T. It is a QMI-internal bug if this property is used in case _device is None. In that
        case, we raise an AssertionError, and we hope the users will complain to us so we can fix the bug in the
        library.

        Raises: AssertionError: in case the property is used when the underlying value of _device is None.

        Returns: The value of _device, if it is not None. """
        assert self._device is not None
        return self._device

    def write(self, data: bytes) -> None:
        self._check_is_open()
        self._safe_device.timeout = self.WRITE_TIMEOUT
        self._safe_device.write_raw(data)

    def _read_message(self, timeout):
        self._safe_device.timeout = timeout
        try:
            data = self._safe_device.read_raw()
        except usb.core.USBError as exc:
            if exc.errno == 110:
                # Timeout
                raise QMI_TimeoutException(str(exc)) from exc
            else:
                raise
        return data

    @staticmethod
    def list_resources():
        resources = usbtmc.list_resources()
        formatted_resources = QMI_UsbTmcTransport._format_resources(resources)
        return formatted_resources



import unittest
from unittest.mock import patch, Mock, MagicMock
from ipaddress import IPv4Address
import urllib.error
import urllib.parse
import urllib.request

import qmi.instruments.aviosys.ippower
from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.instruments.aviosys import Aviosys_IpPower9850, PowerSocket, PowerState


class AviosysIPPower9850ClassTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self._ipv4addess_patch = patch("qmi.instruments.aviosys.ippower.IPv4Address", autospec=IPv4Address)

    def test_instance_creation_with_defaults(self):
        """Test the creation of the IPPower9850 instance with default inputs."""
        # Arrange
        address = "123.123.123.0"
        self._ipv4addess_patch.return_value = address
        expected_baseurl = f"http://{address}"
        expected_username = "admin"
        expected_password = "12345678"
        # Act
        ippower = Aviosys_IpPower9850(QMI_Context("IPPower9850_test"), "ippower", address)
        # Assert
        self.assertEqual(ippower._baseurl, expected_baseurl)
        self.assertEqual(ippower._username, expected_username)
        self.assertEqual(ippower._password, expected_password)

    def test_instance_creation_with_custom_values(self):
        """Test the creation of the IPPower9850 instance with custom inputs."""
        # Arrange
        address = "111.222.210.12"
        self._ipv4addess_patch.return_value = address
        expected_baseurl = f"http://{address}"
        expected_username = "superuser"
        expected_password = "9876543210"
        # Act
        ippower = Aviosys_IpPower9850(
            QMI_Context("IPPower9850_test"), "ippower", address, username=expected_username, password=expected_password)
        # Assert
        self.assertEqual(ippower._baseurl, expected_baseurl)
        self.assertEqual(ippower._username, expected_username)
        self.assertEqual(ippower._password, expected_password)


class AviosysIPPower9850OpenCloseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self._ipv4addess_patch = patch("qmi.instruments.aviosys.ippower.IPv4Address", autospec=IPv4Address)
        self._ipv4addess_patch.return_value = "10.10.10.1"
        self.ippower = Aviosys_IpPower9850(QMI_Context("IPPower9850_test"), "ippower", "10.10.10.1")

    @patch("qmi.instruments.aviosys.ippower.urllib.request", autospec=urllib.request)
    def test_open(self, request_patch):
        """Test the creation of the IPPower9850 instance with default inputs."""
        # Arrange
        expected_baseurl = "http://" + self._ipv4addess_patch.return_value
        request_patch.HTTPBasicAuthHandler = MagicMock()
        request_patch.Request = Mock()
        request_patch.build_opener().open().status = 200
        # Act
        self.ippower.open()
        # Assert
        request_patch.Request.assert_called_once_with(expected_baseurl, method="HEAD")

    @patch("qmi.instruments.aviosys.ippower.urllib.request", autospec=urllib.request)
    def test_open_access_denied(self, request_patch):
        """Test the open fails with "access denied" error."""
        # Arrange
        expected_error = "Access denied for device at {}".format(self._ipv4addess_patch.return_value)
        request_patch.HTTPBasicAuthHandler = MagicMock()
        request_patch.Request = Mock()
        class open_mock:
            def open(self, timeout):
                raise urllib.error.HTTPError("a", 1, "b", Mock(), Mock())

        request_patch.build_opener.return_value = open_mock
        # Act
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.ippower.open()

        # Assert
        self.assertEqual(expected_error, str(exc.exception))

    @patch("qmi.instruments.aviosys.ippower.urllib.request", autospec=urllib.request)
    def test_open_unable_to_find(self, request_patch):
        """Test the open fails with "unable to find" error."""
        # Arrange
        expected_error = "Unable to find a device at {}".format(self._ipv4addess_patch.return_value)
        request_patch.HTTPBasicAuthHandler = MagicMock()
        request_patch.Request = Mock()
        class open_mock:
            def open(self, timeout):
                raise urllib.error.URLError("a", "b")

        request_patch.build_opener.return_value = open_mock
        # Act
        with self.assertRaises(QMI_TimeoutException) as exc:
            self.ippower.open()

        # Assert
        self.assertEqual(expected_error, str(exc.exception))

    @patch("qmi.instruments.aviosys.ippower.urllib.request", autospec=urllib.request)
    def test_open_unknown_error(self, request_patch):
        """Test the open fails with unknown error."""
        # Arrange
        expected_error = "Unknown error in accessing device at {}".format(self._ipv4addess_patch.return_value)
        request_patch.HTTPBasicAuthHandler = MagicMock()
        request_patch.Request = Mock()
        # Act
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.ippower.open()

        # Assert
        self.assertEqual(expected_error, str(exc.exception))

    def test_close(self):
        """Test the close call."""
        # Arrange
        self.ippower._is_open = True  # Make instrument appear as "open"
        # Act
        self.ippower.close()
        # Assert
        self.assertFalse(self.ippower.is_open())

    def test_method_call_excepts_as_opener_is_none(self):
        """Test a method to except due to not having the self._opener yet defined."""
        # Act and Assert
        with self.assertRaises(RuntimeError):
            self.ippower.get_idn()


class AviosysIPPower9850MethodsTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self._ipv4addess_patch = patch("qmi.instruments.aviosys.ippower.IPv4Address", autospec=IPv4Address)
        self._ipv4addess_patch.return_value = "10.10.10.1"
        self.ippower = Aviosys_IpPower9850(QMI_Context("IPPower9850_test"), "ippower", "10.10.10.1")

        with patch("qmi.instruments.aviosys.ippower.urllib.request", autospec=urllib.request) as self.request_patch:
            self.request_patch.HTTPBasicAuthHandler = MagicMock()
            self.request_patch.Request = Mock()
            self.request_patch.build_opener().open().status = 200
            # open the instrument
            self.ippower.open()

    def tearDown(self) -> None:
        self.ippower.close()

    def test_parse_response_excepts(self):
        """Test with a method the _parse_response excepts with invalid response format."""
        # Arrange
        mac = "MAC=1A2B3C4D5E6F7G8H9I0J"
        expected_error = "Invalid response."
        mac_resp = f"{mac}".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(side_effect=[mac_resp])

        # Act
        with self.assertRaises(ValueError) as exc:
            self.ippower.get_idn()

        # Assert
        self.assertEqual(expected_error, str(exc.exception))

    def test_parse_status_string_excepts(self):
        """Test the _parse_status_string excepts at invalid response."""
        # Arrange
        expected_error = "Instrument status string is invalid."
        states = "p61=2p62=3p63=4p64=5"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{states}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(return_value=state_resp)

        # Act
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.ippower.get_all_states()

        # Assert
        self.assertEqual(expected_error, str(exc.exception))

    def test_send_command_excepts(self):
        """Test with a method the _send_command excepts with HTTPError."""
        # Arrange
        expected_error = "Error in communication with device"
        self.request_patch.build_opener().open().__enter__().read = Mock(
            side_effect=[urllib.error.HTTPError("a", 1, "b", Mock(), Mock())]
        )

        # Act
        with self.assertRaises(QMI_InstrumentException) as exc:
            self.ippower.get_idn()

        # Assert
        self.assertEqual(expected_error, str(exc.exception))

    def test_get_idn(self):
        """Test the get_idn method."""
        # Arrange
        mac = "MAC=1A2B3C4D5E6F7G8H9I0J"
        version = "Version=1.23.4"
        expected_vendor = "Aviosys"
        expected_model = "IP Power 9850XX"
        expected_serial = ":".join(mac[i:i+2] for i in range(4, 16, 2))
        expected_version = version.split("=")[1]
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        mac_resp = f"<!--CGI-DATABEG-->\n<p>\n{mac}</p>\n\n<!--CGI-DATAEND-->".encode()
        version_resp = f"<!--CGI-DATABEG-->\n<p>\n{version}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(side_effect=[mac_resp, version_resp])

        # Act
        idn = self.ippower.get_idn()

        # Assert
        self.assertEqual(expected_vendor, idn.vendor)
        self.assertEqual(expected_model, idn.model)
        self.assertEqual(expected_serial, idn.serial)
        self.assertEqual(expected_version, idn.version)

    def test_get_all_states(self):
        """Test the get_all_states method."""
        # Arrange
        expected_P61 = PowerState.OFF
        expected_P62 = PowerState.OFF
        expected_P63 = PowerState.ON
        expected_P64 = PowerState.OFF
        states = f"p61={expected_P61.value}p62={expected_P62.value}p63={expected_P63.value}p64={expected_P64.value}"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{states}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(return_value=state_resp)

        # Act
        resp = self.ippower.get_all_states()

        # Assert
        self.assertEqual(resp[PowerSocket.P1], expected_P61)
        self.assertEqual(resp[PowerSocket.P2], expected_P62)
        self.assertEqual(resp[PowerSocket.P3], expected_P63)
        self.assertEqual(resp[PowerSocket.P4], expected_P64)

    def test_get_state(self):
        """Test the get_state method."""
        # Arrange
        expected_P61 = PowerState.OFF
        expected_P62 = PowerState.OFF
        expected_P63 = PowerState.ON
        expected_P64 = PowerState.OFF
        states = f"p61={expected_P61.value}p62={expected_P62.value}p63={expected_P63.value}p64={expected_P64.value}"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{states}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(return_value=state_resp)

        # Act
        resp_P61 = self.ippower.get_state(PowerSocket.P1)
        resp_P62 = self.ippower.get_state(PowerSocket.P2)
        resp_P63 = self.ippower.get_state(PowerSocket.P3)
        resp_P64 = self.ippower.get_state(PowerSocket.P4)

        # Assert
        self.assertEqual(resp_P61, expected_P61)
        self.assertEqual(resp_P62, expected_P62)
        self.assertEqual(resp_P63, expected_P63)
        self.assertEqual(resp_P64, expected_P64)

    def test_set_state(self):
        """Test the set_state method."""
        # Arrange
        expected_P61 = PowerState.OFF
        expected_P62 = PowerState.OFF
        expected_P63 = PowerState.ON
        expected_P64 = PowerState.OFF
        states = f"p61={expected_P61.value}p62={expected_P62.value}p63={expected_P63.value}p64={expected_P64.value}"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{states}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(return_value=state_resp)

        # Act
        resp_P61 = self.ippower.set_state(PowerSocket.P1, expected_P61)
        resp_P62 = self.ippower.set_state(PowerSocket.P2, expected_P62)
        resp_P63 = self.ippower.set_state(PowerSocket.P3, expected_P63)
        resp_P64 = self.ippower.set_state(PowerSocket.P4, expected_P64)

        # Assert
        self.assertEqual(resp_P61, bool(expected_P61))
        self.assertEqual(resp_P62, bool(expected_P62))
        self.assertEqual(resp_P63, bool(expected_P63))
        self.assertEqual(resp_P64, bool(expected_P64))

    def test_set_states_true(self):
        """Test the set_states method."""
        # Arrange
        expected_P61 = PowerState.OFF
        expected_P62 = PowerState.OFF
        expected_P63 = PowerState.ON
        expected_P64 = PowerState.OFF
        set_states = {
            PowerSocket.P1: expected_P61,
            PowerSocket.P2: expected_P62,
            PowerSocket.P3: expected_P63,
            PowerSocket.P4: expected_P64,
        }
        states = f"p61={expected_P61.value}p62={expected_P62.value}p63={expected_P63.value}p64={expected_P64.value}"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{states}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(return_value=state_resp)

        # Act
        resp = self.ippower.set_states(set_states)

        # Assert
        self.assertTrue(resp)

    def test_set_states_false(self):
        """Test the set_states method."""
        # Arrange
        expected_P61 = PowerState.OFF
        expected_P62 = PowerState.OFF
        expected_P63 = PowerState.ON
        expected_P64 = PowerState.OFF
        set_states = {
            PowerSocket.P1: expected_P61,
            PowerSocket.P2: PowerState.ON,
            PowerSocket.P3: expected_P63,
            PowerSocket.P4: expected_P64,
        }
        states = f"p61={expected_P61.value}p62={expected_P62.value}p63={expected_P63.value}p64={expected_P64.value}"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{states}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(return_value=state_resp)

        # Act
        resp = self.ippower.set_states(set_states)

        # Assert
        self.assertFalse(resp)

    def test_set_all_off(self):
        """Test the set_all_off method."""
        # Arrange
        states = "p61=0p62=0p63=0p64=0"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{states}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(return_value=state_resp)

        # Act
        resp = self.ippower.set_all_off()

        # Assert
        self.assertTrue(resp)

    def test_set_all_on(self):
        """Test the set_all_on method."""
        # Arrange
        states = "p61=1p62=1p63=1p64=1"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{states}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(return_value=state_resp)

        # Act
        resp = self.ippower.set_all_on()

        # Assert
        self.assertTrue(resp)

    def test_cycle(self):
        """Test the cycle method."""
        # Arrange
        state_strings = "p61 cycle ok"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{state_strings}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(return_value=state_resp)

        # Act
        resp = self.ippower.cycle(PowerSocket.P1)

        # Assert
        self.assertTrue(resp)

    def test_cycle_with_block_and_wait(self):
        """Test the cycle method with block and wait input parameters."""
        # Arrange
        state_strings = "p61 cycle ok"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{state_strings}</p>\n\n<!--CGI-DATAEND-->".encode()
        states_0 = "<!--CGI-DATABEG-->\n<p>\np61=0p62=0p63=0p64=0</p>\n\n<!--CGI-DATAEND-->".encode()
        states_1 = "<!--CGI-DATABEG-->\n<p>\np61=1p62=0p63=0p64=0</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(side_effect=[state_resp, states_0, states_1])

        # Act
        resp = self.ippower.cycle(PowerSocket.P1, wait=0.01, block=True)

        # Assert
        self.assertTrue(resp)

    def test_cycle_with_block_and_wait_cycle_nok(self):
        """Test the cycle method with block and wait input parameters, with nok cycle."""
        # Arrange
        qmi.instruments.aviosys.ippower.TIMEOUT = 1.0
        state_strings = "p61 cycle nok"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{state_strings}</p>\n\n<!--CGI-DATAEND-->".encode()
        states = "<!--CGI-DATABEG-->\n<p>\np61=0p62=0p63=0p64=0</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(side_effect=[state_resp, states])

        # Act
        resp = self.ippower.cycle(PowerSocket.P1, wait=0.01, block=True)

        # Assert
        self.assertFalse(resp)

    def test_cycle_excepts(self):
        """Test the cycle method excepts on invalid status string."""
        # Arrange
        qmi.instruments.aviosys.ippower.TIMEOUT = 1.0
        state_strings = "p65 cycle ok"
        # Response is always in the form:
        #     <!--CGI-DATABEG-->\n<p>\n...</p>\n\n<!--CGI-DATAEND-->
        state_resp = f"<!--CGI-DATABEG-->\n<p>\n{state_strings}</p>\n\n<!--CGI-DATAEND-->".encode()
        self.request_patch.build_opener().open().__enter__().read = Mock(return_value=state_resp)

        # Act and Assert
        with self.assertRaises(QMI_InstrumentException):
            self.ippower.cycle(PowerSocket.P1)


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python3

"""Unit tests for ADbasic parser."""
import logging
import os.path
import tempfile
import argparse
import unittest
from unittest import mock
import io
import contextlib

from qmi.utils.adbasic_parser import (
    ParDesc, FParDesc, ArrayElemDesc, ParseException, SymbolInfo,
    parse_adbasic_program, analyze_parameter_info, main)


TEST_PROGRAM_FILE = "program.bas"
TEST_PROGRAM = r"""' Test program for ADbasic parser.

#Include ADwinPro_All.inc
#Include .\testprog.inc  ' this include file will also be parsed

' Some misc defines
#Define Pi 3.14159
#DEFINE Symbol_name Value

' Define named parameters.
#Define PAR_one Par_1
#Define PAR_float FPar_1

' Define data arrays.
#Define DATA_params Data_10

' Define parameters located in array elements.
#Define PAR_elem_one DATA_params[1]
#Define PAR_elem_three DATA_params[3]

Dim Data_10[5] As Long

Init:

Event:
"""

TEST_PROGRAM_NO_INCLUDES_FILE = "program_noinc.bas"
TEST_PROGRAM_NO_INCLUDES = r"""' Test program for ADbasic parser.

#Include /  ' this missing include file should cause logger warning

' Some misc defines
#DEFINE Symbol_name Value
#Define Pi 3.14159

Init:

Event:
"""

TEST_INCLUDE_FILE = "testprog.inc"
TEST_INCLUDE = r"""' Test include file for ADbasic parser.

#define testsym 31
#Define PAR_TWO PAR_2
#Define DATA_result Data_11
"""


def write_source_files(tempdir):
    with open(os.path.join(tempdir, TEST_PROGRAM_FILE), "w") as outf:
        outf.write(TEST_PROGRAM)
    with open(os.path.join(tempdir, TEST_INCLUDE_FILE), "w") as outf:
        outf.write(TEST_INCLUDE)


def write_source_file(tempdir, fn, content):
    with open(os.path.join(tempdir, fn), "w") as outf:
        outf.write(content)


class TestAdbasicParser(unittest.TestCase):

    def test_parse_symbols(self):

        with tempfile.TemporaryDirectory(prefix="qmitest") as tempdir:
            write_source_files(tempdir)
            filename = os.path.join(tempdir, TEST_PROGRAM_FILE)
            symbols = parse_adbasic_program(filename, tempdir)

        expect_symbols = [
            (TEST_PROGRAM_FILE, 7, "Pi", "3.14159"),
            (TEST_PROGRAM_FILE, 8, "Symbol_name", "Value"),
            (TEST_PROGRAM_FILE, None, "PAR_one", "Par_1"),
            (TEST_PROGRAM_FILE, None, "PAR_float", "FPar_1"),
            (TEST_PROGRAM_FILE, None, "DATA_params", "Data_10"),
            (TEST_PROGRAM_FILE, None, "PAR_elem_one", "DATA_params[1]"),
            (TEST_PROGRAM_FILE, None, "PAR_elem_three", "DATA_params[3]"),
            (TEST_INCLUDE_FILE, 3, "testsym", "31"),
            (TEST_INCLUDE_FILE, None, "PAR_TWO", "PAR_2"),
            (TEST_INCLUDE_FILE, None, "DATA_result", "Data_11")
        ]

        self.assertEqual(len(symbols), len(expect_symbols))

        for (symbol, expect) in zip(symbols, expect_symbols):
            file_path = symbol.filename.replace("\\", "/").replace("/./", "/")
            expect_dir = tempdir.replace("\\", "/") + "/"
            self.assertEqual(file_path, expect_dir + expect[0])
            if expect[1] is not None:
                self.assertEqual(symbol.line_nr, expect[1])

            self.assertEqual(symbol.label, expect[2])
            self.assertEqual(symbol.value, expect[3])

    def test_parse_symbols_no_dots(self):
        """Test the path is returned into a subfolder if no dots present."""
        global TEST_PROGRAM
        with tempfile.TemporaryDirectory(prefix="qmitest") as tempdir:
            TEST_PROGRAM = TEST_PROGRAM.replace(".\\", "subfolder\\")
            write_source_files(tempdir)
            subdir = os.path.join(tempdir, "subfolder")
            os.mkdir(subdir)
            write_source_file(subdir, TEST_INCLUDE_FILE, TEST_INCLUDE)
            filename = os.path.join(tempdir, TEST_PROGRAM_FILE)
            symbols = parse_adbasic_program(filename, tempdir)

        expect_symbols = [
            (TEST_PROGRAM_FILE, 7, "Pi", "3.14159"),
            (TEST_PROGRAM_FILE, 8, "Symbol_name", "Value"),
            (TEST_PROGRAM_FILE, None, "PAR_one", "Par_1"),
            (TEST_PROGRAM_FILE, None, "PAR_float", "FPar_1"),
            (TEST_PROGRAM_FILE, None, "DATA_params", "Data_10"),
            (TEST_PROGRAM_FILE, None, "PAR_elem_one", "DATA_params[1]"),
            (TEST_PROGRAM_FILE, None, "PAR_elem_three", "DATA_params[3]"),
            ("subfolder/" + TEST_INCLUDE_FILE, 3, "testsym", "31"),
            ("subfolder/" + TEST_INCLUDE_FILE, None, "PAR_TWO", "PAR_2"),
            ("subfolder/" + TEST_INCLUDE_FILE, None, "DATA_result", "Data_11")
        ]
        self.assertEqual(len(symbols), len(expect_symbols))

        for (symbol, expect) in zip(symbols, expect_symbols):
            file_path = symbol.filename.replace("\\", "/").replace("/./", "/")
            expect_dir = tempdir.replace("\\", "/") + "/"
            self.assertEqual(file_path, expect_dir + expect[0])
            if expect[1] is not None:
                self.assertEqual(symbol.line_nr, expect[1])

            self.assertEqual(symbol.label, expect[2])
            self.assertEqual(symbol.value, expect[3])

    def test_param_info(self):
        # Arrange
        expect_params = [
            ("one", ParDesc, {"par_index": 1}),
            ("TWO", ParDesc, {"par_index": 2}),
            ("float", FParDesc, {"fpar_index": 1}),
            ("elem_one", ArrayElemDesc, {"data_index": 10, "elem_index": 1}),
            ("elem_three", ArrayElemDesc, {"data_index": 10, "elem_index": 3})
        ]

        expect_data = [
            ("params", 10),
            ("result", 11)
        ]
        with tempfile.TemporaryDirectory(prefix="qmitest") as tempdir:
            write_source_files(tempdir)
            filename = os.path.join(tempdir, TEST_PROGRAM_FILE)
            symbols = parse_adbasic_program(filename, tempdir)

        # Act
        param_info = analyze_parameter_info(symbols)

        # Assert
        self.assertEqual(str(param_info.param["one"]), "Par_1")
        self.assertEqual(str(param_info.param["float"]), "FPar_1")
        self.assertEqual(str(param_info.param["elem_one"]), "Data_10[1]")
        self.assertEqual(len(param_info.param), len(expect_params))
        for (name, cls, args) in expect_params:
            desc = param_info.param[name]
            self.assertIsInstance(desc, cls)
            for (arg_name, arg_val) in args.items():
                self.assertEqual(getattr(desc, arg_name), arg_val)

        self.assertEqual(len(param_info.data), len(expect_data))
        for (name, index) in expect_data:
            self.assertEqual(param_info.data[name], index)

    def test_wrong_symbol_definition_format(self):
        """Test wrong symbol definition format"""

        logging.getLogger("qmi.utils.adbasic_parser").setLevel(logging.CRITICAL)
        wrong_symbol_info = SymbolInfo(
            filename="program.bas",
            line_nr=1,
            label="DATA_params",
            value="Date_10"
        )
        param_info = analyze_parameter_info([wrong_symbol_info])
        self.assertEqual(0, len(param_info.param))
        self.assertEqual(0, len(param_info.data))

    def test_duplicate_case(self):
        """Test duplicate definition of this symbol with different case"""
        expected_exc = "program.bas, line 2: Duplicate definition of symbol DATA_Params with different case"
        wrong_symbol_info = [SymbolInfo(
            filename="program.bas",
            line_nr=1,
            label="DATA_params",
            value="Data_10"
            ),
            SymbolInfo(
                filename="program.bas",
                line_nr=2,
                label="DATA_Params",
                value="Data_10"
            )
        ]
        with self.assertRaises(ParseException) as exc:
            analyze_parameter_info(wrong_symbol_info)

        self.assertEqual(str(exc.exception), expected_exc)

    def test_duplicate_index(self):
        """Test duplicate definition of this symbol with different index"""
        expected_exc = "program.bas, line 1: Duplicate definition of symbol DATA_params for different data array"
        wrong_symbol_info = [SymbolInfo(
            filename="program.bas",
            line_nr=1,
            label="DATA_params",
            value="Data_10"
            ),
            SymbolInfo(
                filename="program.bas",
                line_nr=1,
                label="DATA_params",
                value="Data_9"
            )
        ]
        with self.assertRaises(ParseException) as exc:
            analyze_parameter_info(wrong_symbol_info)

        self.assertEqual(str(exc.exception), expected_exc)

    def test_duplicate_use(self):
        """Test duplicate use of the same Data_nnn array under different names"""
        expected_exc = "program.bas, line 1: Symbol DATA_rapams is a duplicate reference to Data_10"
        wrong_symbol_info = [SymbolInfo(
            filename="program.bas",
            line_nr=1,
            label="DATA_params",
            value="Data_10"
            ),
            SymbolInfo(
                filename="program.bas",
                line_nr=1,
                label="DATA_rapams",
                value="Data_10"
            )
        ]
        with self.assertRaises(ParseException) as exc:
            analyze_parameter_info(wrong_symbol_info)

        self.assertEqual(str(exc.exception), expected_exc)


class TestParseException(unittest.TestCase):

    def test_init(self):
        """Test that __init__ works"""
        # Arrange
        filename = "foo.bar"
        line_nr = 42
        message = "barfood"
        expected_string = "{}, line {}: {}".format(filename, line_nr, message)

        # Act
        parse_exception_instance = ParseException(filename, line_nr, message)

        # Assert
        self.assertEqual(str(parse_exception_instance), expected_string)


class TestParseAdbasicProgram(unittest.TestCase):

    def test_parse_adbasic_program_empty_filename_raises_exception(self):
        """See that exception is raised with empty file name"""
        with self.assertRaises(FileNotFoundError):
            parse_adbasic_program("", "")
            

class TestMain(unittest.TestCase):

    def setUp(self) -> None:
        global TEST_PROGRAM
        # Change back
        TEST_PROGRAM = TEST_PROGRAM.replace("subfolder\\", ".\\")
        self.tempdir = tempfile.TemporaryDirectory(prefix="qmitest")
        write_source_files(self.tempdir.name)
        self.filename = os.path.join(self.tempdir.name, TEST_PROGRAM_FILE)

    def tearDown(self) -> None:
        self.tempdir.cleanup()

    def test_main(self):
        """Test that the main function parses the input files correctly and the print statement prints symbols of type
        symbol_name only (so, not symbols without '_')."""
        # Arrange
        expected = [
        "PAR_{:32} = Par_1".format("one"),
        "PAR_{:32} = Par_2".format("TWO"),
        "PAR_{:32} = FPar_1".format("float"),
        "PAR_{:32} = Data_10[1]".format("elem_one"),
        "PAR_{:32} = Data_10[3]".format("elem_three"),
        "DATA_{:32} = Data_10".format("params"),
        "DATA_{:32} = Data_11".format("result"),
        ""
        ]
        with mock.patch("argparse.ArgumentParser.parse_args", return_value=argparse.Namespace(filename=self.filename)):
            # Act
            with io.StringIO() as buf:
                with contextlib.redirect_stdout(buf):
                    main()

                stdout_result = buf.getvalue()

        # Assert
        self.assertListEqual(stdout_result.split("\n"), expected)


class TestNoIncludeFilesMain(unittest.TestCase):

    def setUp(self) -> None:
        self.tempdir = tempfile.TemporaryDirectory(prefix="qmitest")
        write_source_file(self.tempdir.name, TEST_PROGRAM_NO_INCLUDES_FILE, TEST_PROGRAM_NO_INCLUDES)
        self.filename = os.path.join(self.tempdir.name, TEST_PROGRAM_NO_INCLUDES_FILE)

    def tearDown(self) -> None:
        self.tempdir.cleanup()

    def test_main(self):
        """Test that no include file creates a logger warning"""
        # Arrange
        with mock.patch("argparse.ArgumentParser.parse_args", return_value=argparse.Namespace(filename=self.filename)):
            with mock.patch("logging.Logger.warning") as warn:
                # Act
                main()
                # Assert
                warn.assert_called_once_with("Absolute include path not supported: %r", r"/")


if __name__ == "__main__":
    unittest.main()



import ctypes
import sys
import time
from typing import Any

import unittest
from unittest.mock import patch, call

from qmi.core.exceptions import QMI_InstrumentException, QMI_InvalidOperationException
from qmi.instruments.picoquant._picoquant import _PicoquantHarp

from tests.patcher import PatcherQmiContext as QMI_Context


_shlib_function_signatures = [
        ('SH_GetLibraryVersion', ctypes.c_int, [('vers', ctypes.POINTER(ctypes.c_char))]),
        ('SH_GetErrorString', ctypes.c_int, [('errstring', ctypes.POINTER(ctypes.c_char)), ('errcode', ctypes.c_int)]),
        ('SH_OpenDevice', ctypes.c_int, [('devidx', ctypes.c_int), ('serial', ctypes.POINTER(ctypes.c_char))]),
        ('SH_CloseDevice', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('SH_Initialize', ctypes.c_int, [('devidx', ctypes.c_int), ('mode', ctypes.c_int), ('refsource', ctypes.c_int)]),
        ('SH_GetHardwareInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('model', ctypes.POINTER(ctypes.c_char)), ('partno', ctypes.POINTER(ctypes.c_char)), ('version', ctypes.POINTER(ctypes.c_char))]),
        ('SH_GetSerialNumber', ctypes.c_int, [('devidx', ctypes.c_int), ('serial', ctypes.POINTER(ctypes.c_char))]),
        ('SH_GetFeatures', ctypes.c_int, [('devidx', ctypes.c_int), ('features', ctypes.POINTER(ctypes.c_int))]),
        ('SH_GetBaseResolution', ctypes.c_int, [('devidx', ctypes.c_int), ('resolution', ctypes.POINTER(ctypes.c_double)), ('binsteps', ctypes.POINTER(ctypes.c_int))]),
        ('SH_GetNumOfInputChannels', ctypes.c_int, [('devidx', ctypes.c_int), ('nchannels', ctypes.POINTER(ctypes.c_int))]),
        ('SH_SetSyncDiv', ctypes.c_int, [('devidx', ctypes.c_int), ('div', ctypes.c_int)]),
        ('SH_SetSyncEdgeTrg', ctypes.c_int, [('devidx', ctypes.c_int), ('level', ctypes.c_int), ('edge', ctypes.c_int)]),
        ('SH_SetSyncChannelOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('value', ctypes.c_int)]),
        ('SH_SetSyncDeadTime', ctypes.c_int, [('devidx', ctypes.c_int), ('on', ctypes.c_int), ('deadtime', ctypes.c_int)]),
        ('SH_SetInputEdgeTrg', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('level', ctypes.c_int), ('edge', ctypes.c_int)]),
        ('SH_SetInputChannelOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('value', ctypes.c_int)]),
        ('SH_SetInputDeadTime', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('on', ctypes.c_int), ('deadtime', ctypes.c_int)]),
        ('SH_SetInputChannelEnable', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('enable', ctypes.c_int)]),
        ('SH_SetStopOverflow', ctypes.c_int, [('devidx', ctypes.c_int), ('stop_ovfl', ctypes.c_int), ('stopcount', ctypes.c_uint)]),
        ('SH_SetBinning', ctypes.c_int, [('devidx', ctypes.c_int), ('binning', ctypes.c_int)]),
        ('SH_SetOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('offset', ctypes.c_int)]),
        ('SH_SetHistoLen', ctypes.c_int, [('devidx', ctypes.c_int), ('lencode', ctypes.c_int), ('actuallen', ctypes.POINTER(ctypes.c_int))]),
        ('SH_SetMeasControl', ctypes.c_int, [('devidx', ctypes.c_int), ('control', ctypes.c_int), ('startedge', ctypes.c_int), ('stopedge', ctypes.c_int)]),
        ('SH_SetTriggerOutput', ctypes.c_int, [('devidx', ctypes.c_int), ('period', ctypes.c_int)]),
        ('SH_ClearHistMem', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('SH_StartMeas', ctypes.c_int, [('devidx', ctypes.c_int), ('tacq', ctypes.c_int)]),
        ('SH_StopMeas', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('SH_CTCStatus', ctypes.c_int, [('devidx', ctypes.c_int), ('ctcstatus', ctypes.POINTER(ctypes.c_int))]),
        ('SH_GetHistogram', ctypes.c_int, [('devidx', ctypes.c_int), ('chcount', ctypes.POINTER(ctypes.c_uint)), ('channel', ctypes.c_int)]),
        ('SH_GetAllHistograms', ctypes.c_int, [('devidx', ctypes.c_int), ('chcount', ctypes.POINTER(ctypes.c_uint))]),
        ('SH_GetResolution', ctypes.c_int, [('devidx', ctypes.c_int), ('resolution', ctypes.POINTER(ctypes.c_double))]),
        ('SH_GetSyncPeriod', ctypes.c_int, [('devidx', ctypes.c_int), ('period', ctypes.POINTER(ctypes.c_double))]),
        ('SH_GetSyncRate', ctypes.c_int, [('devidx', ctypes.c_int), ('syncrate', ctypes.POINTER(ctypes.c_int))]),
        ('SH_GetCountRate', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('cntrate', ctypes.POINTER(ctypes.c_int))]),
        ('SH_GetAllCountRates', ctypes.c_int, [('devidx', ctypes.c_int), ('syncrate', ctypes.POINTER(ctypes.c_int)), ('cntrates', ctypes.POINTER(ctypes.c_int))]),
        ('SH_GetFlags', ctypes.c_int, [('devidx', ctypes.c_int), ('flags', ctypes.POINTER(ctypes.c_int))]),
        ('SH_GetElapsedMeasTime', ctypes.c_int, [('devidx', ctypes.c_int), ('elapsed', ctypes.POINTER(ctypes.c_double))]),
        ('SH_GetStartTime', ctypes.c_int, [('devidx', ctypes.c_int), ('timedw2', ctypes.POINTER(ctypes.c_uint)), ('timedw1', ctypes.POINTER(ctypes.c_uint)), ('timedw0', ctypes.POINTER(ctypes.c_uint))]),
        ('SH_GetWarnings', ctypes.c_int, [('devidx', ctypes.c_int), ('warnings', ctypes.POINTER(ctypes.c_int))]),
        ('SH_GetWarningsText', ctypes.c_int, [('devidx', ctypes.c_int), ('text', ctypes.POINTER(ctypes.c_char)), ('warnings', ctypes.c_int)]),
        ('SH_SetMarkerHoldoffTime', ctypes.c_int, [('devidx', ctypes.c_int), ('holdofftime', ctypes.c_int)]),
        ('SH_SetMarkerEdges', ctypes.c_int, [('devidx', ctypes.c_int), ('me1', ctypes.c_int), ('me2', ctypes.c_int), ('me3', ctypes.c_int), ('me4', ctypes.c_int)]),
        ('SH_SetMarkerEnable', ctypes.c_int, [('devidx', ctypes.c_int), ('en1', ctypes.c_int), ('en2', ctypes.c_int), ('en3', ctypes.c_int), ('en4', ctypes.c_int)]),
        ('SH_ReadFiFo', ctypes.c_int, [('devidx', ctypes.c_int), ('buffer', ctypes.POINTER(ctypes.c_uint)), ('nactual', ctypes.POINTER(ctypes.c_int))]),
        ('SH_GetDebugInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('debuginfo', ctypes.POINTER(ctypes.c_char))]),
        ('SH_GetNumOfModules', ctypes.c_int, [('devidx', ctypes.c_int), ('nummod', ctypes.POINTER(ctypes.c_int))]),
        ('SH_GetModuleInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('modidx', ctypes.c_int), ('modelcode', ctypes.POINTER(ctypes.c_int)), ('versioncode', ctypes.POINTER(ctypes.c_int))]),
        ('SH_WRabbitGetMAC', ctypes.c_int, [('devidx', ctypes.c_int), ('mac_addr', ctypes.POINTER(ctypes.c_char))]),
        ('SH_WRabbitSetMAC', ctypes.c_int, [('devidx', ctypes.c_int), ('mac_addr', ctypes.POINTER(ctypes.c_char))]),
        ('SH_WRabbitGetInitScript', ctypes.c_int, [('devidx', ctypes.c_int), ('initscript', ctypes.POINTER(ctypes.c_char))]),
        ('SH_WRabbitSetInitScript', ctypes.c_int, [('devidx', ctypes.c_int), ('initscript', ctypes.POINTER(ctypes.c_char))]),
        ('SH_WRabbitGetSFPData', ctypes.c_int, [('devidx', ctypes.c_int), ('sfpnames', ctypes.POINTER(ctypes.c_char)), ('dTxs', ctypes.POINTER(ctypes.c_int)), ('dRxs', ctypes.POINTER(ctypes.c_int)), ('alphas', ctypes.POINTER(ctypes.c_int))]),
        ('SH_WRabbitSetSFPData', ctypes.c_int, [('devidx', ctypes.c_int), ('sfpnames', ctypes.POINTER(ctypes.c_char)), ('dTxs', ctypes.POINTER(ctypes.c_int)), ('dRxs', ctypes.POINTER(ctypes.c_int)), ('alphas', ctypes.POINTER(ctypes.c_int))]),
        ('SH_WRabbitInitLink', ctypes.c_int, [('devidx', ctypes.c_int), ('link_on', ctypes.c_int)]),
        ('SH_WRabbitSetMode', ctypes.c_int, [('devidx', ctypes.c_int), ('bootfromscript', ctypes.c_int), ('reinit_with_mode', ctypes.c_int), ('mode', ctypes.c_int)]),
        ('SH_WRabbitSetTime', ctypes.c_int, [('devidx', ctypes.c_int), ('timehidw', ctypes.c_uint), ('timelodw', ctypes.c_uint)]),
        ('SH_WRabbitGetTime', ctypes.c_int, [('devidx', ctypes.c_int), ('timehidw', ctypes.POINTER(ctypes.c_uint)), ('timelodw', ctypes.POINTER(ctypes.c_uint)), ('subsec16ns', ctypes.POINTER(ctypes.c_uint))]),
        ('SH_WRabbitGetStatus', ctypes.c_int, [('devidx', ctypes.c_int), ('wrstatus', ctypes.POINTER(ctypes.c_int))]),
        ('SH_WRabbitGetTermOutput', ctypes.c_int, [('devidx', ctypes.c_int), ('buffer', ctypes.POINTER(ctypes.c_char)), ('nchar', ctypes.POINTER(ctypes.c_int))])
    ]


class _LibWrapper:

    def __init__(self, t: str):

        if not sys.platform.startswith("linux"):
            raise ValueError("Unsupported platform.")

        if t == "SH":
            self._lib = ctypes.cdll.LoadLibrary("libsh000.so")
            self.annotate_function_signatures(_shlib_function_signatures)
            self._prefix = "SH"
        else:
            raise ValueError("Unknown library: {lib}.".format(lib=t))

    def __getattr__(self, item: str) -> Any:
        attr_name = "{prefix}_{item}".format(prefix=self._prefix, item=item)
        attr = getattr(self._lib, attr_name)

        def wrap_fun(*args, **kwargs):
            errcode = attr(*args, **kwargs)
            if errcode != 0:
                raise QMI_InstrumentException(f"Interaction with PicoQuant library failed, errorcode [{errcode}].")
        return wrap_fun

    def annotate_function_signatures(self, sigs) -> None:
        """Annotate functions present in the MultiHarp shared library according to their function signatures."""
        function_signatures = sigs

        for (name, restype, argtypes) in function_signatures:
            try:
                func = getattr(self._lib, name)
                func.restype = restype
                func.argtypes = [argtype for (argname, argtype) in argtypes]
            except AttributeError:
                # Ignore functions that cannot be found.
                pass


class PicoQuant_SomeHarp000(_PicoquantHarp):
    """Instrument driver for the PicoQuant SomeHarp 000."""

    _MODEL = "SH"
    _MAXDEVNUM = 8
    _TTREADMAX = 131072

    @property
    def _max_dev_num(self):
        return self._MAXDEVNUM

    @property
    def _ttreadmax(self):
        return self._TTREADMAX

    @property
    def _lib(self) -> _LibWrapper:
        if self._lazy_lib is None:
            self._lazy_lib = _LibWrapper('SH')
        return self._lazy_lib

    @property
    def _model(self):
        return self._MODEL


class PicoQuantSomeHarpOpenTestCase(unittest.TestCase):

    def setUp(self) -> None:
        patcher = patch('ctypes.cdll.LoadLibrary', spec=ctypes.CDLL)
        self._library_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)

    def test_open_close(self):
        """Test regular open where SN is returned immediately."""
        someharp = PicoQuant_SomeHarp000(QMI_Context("test_someharp"), 'someharp', '1111111')

        self._library_mock.SH_GetLibraryVersion.return_value = 0
        self._library_mock.SH_OpenDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(8)
        string_buffer.value = b'1111111'

        with patch('sys.platform', 'linux1'):
            with patch('ctypes.create_string_buffer', return_value=string_buffer):
                someharp.open()

            someharp.close()

    def test_open_model_not_implemented_exception(self):
        """Test that model type 'SH' raises an exception if no device serial number is immediately received."""
        someharp = PicoQuant_SomeHarp000(QMI_Context("test_someharp"), 'someharp', '1111111')

        self._library_mock.SH_GetLibraryVersion.return_value = 0
        self._library_mock.SH_OpenDevice.return_value = 0
        self._library_mock.SH_CloseDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(8)
        string_buffer.value = b''

        with patch('sys.platform', 'linux1'):
            with patch('ctypes.create_string_buffer', return_value=string_buffer):
                with self.assertRaises(NotImplementedError):
                    someharp.open()

    def test_open_excepts_wrong_serial(self):
        """When trying to find an instrument with a wrong serial, we fail after _MAXDEVNUM SH_OpenDevice calls."""
        wrong_serial = '1111111'
        serial = '1111112'
        someharp = PicoQuant_SomeHarp000(QMI_Context("test_someharp"), 'someharp', serial)

        self._library_mock.SH_GetLibraryVersion.return_value = 0
        self._library_mock.SH_OpenDevice.return_value = 0
        self._library_mock.SH_CloseDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(8)
        string_buffer.value = bytes(wrong_serial.encode())
        expected_exception = f"No device with serial number {serial!r} found."
        expected_calls = [call(i, string_buffer) for i in range(PicoQuant_SomeHarp000._MAXDEVNUM)]

        with patch('sys.platform', 'linux1'), self.assertRaises(QMI_InstrumentException) as exc:
            with patch('ctypes.create_string_buffer', return_value=string_buffer):
                someharp.open()

            self._library_mock.SH_OpenDevice.assert_has_calls(expected_calls)
            self.assertEqual(expected_exception, str(exc.exception))


class PicoQuantSomeHarpTestCase(unittest.TestCase):

    def setUp(self) -> None:
        patcher = patch('ctypes.cdll.LoadLibrary', spec=ctypes.CDLL)
        self._library_mock = patcher.start().return_value

        function_names, _, _ = zip(*_shlib_function_signatures)
        self._library_mock.mock_add_spec(function_names, spec_set=True)

        self.addCleanup(patcher.stop)

        self._someharp = PicoQuant_SomeHarp000(QMI_Context("test_someharp"), 'someharp', '1111111')

        self._library_mock.SH_GetLibraryVersion.return_value = 0
        self._library_mock.SH_OpenDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(b'1111111')

        with patch('sys.platform', 'linux1'), patch('ctypes.create_string_buffer', return_value=string_buffer):
            self._someharp.open()

    def tearDown(self) -> None:
        self._library_mock.SH_CloseDevice.return_value = 0
        self._someharp.close()

    def test_get_error_string(self):
        """Test that get_error_string can be called and returns a string."""
        self._library_mock.SH_GetErrorString.return_value = 0
        expected_error_string = "Device is busy."

        error_code = -1
        string_buffer = ctypes.create_string_buffer(expected_error_string.encode('ASCII'))

        with patch('ctypes.create_string_buffer', return_value=string_buffer):
            error_string = self._someharp.get_error_string(error_code)

        self._library_mock.SH_GetErrorString.assert_called_once_with(string_buffer, error_code)
        self.assertEqual(error_string, expected_error_string)

    def test_get_hardware_info(self):

        self._library_mock.SH_GetHardwareInfo.return_value = 0
        expected_hardware_info = 'abcdefgh', '12345678', '23456789'

        model_string_buffer = ctypes.create_string_buffer(expected_hardware_info[0].encode('ASCII'), 24)
        partno = ctypes.create_string_buffer(expected_hardware_info[1].encode('ASCII'), 8)
        version = ctypes.create_string_buffer(expected_hardware_info[2].encode('ASCII'), 8)

        with patch('ctypes.create_string_buffer', side_effect=[model_string_buffer, partno, version]):
            hardware_info = self._someharp.get_hardware_info()

        self.assertTupleEqual(hardware_info, expected_hardware_info)

    def test_get_serial_number(self):
        self._library_mock.SH_GetSerialNumber.return_value = 0
        expected_serial_number = "1111111"

        string_buffer = ctypes.create_string_buffer(expected_serial_number.encode('ASCII'))

        with patch('ctypes.create_string_buffer', return_value=string_buffer):
            serial_number = self._someharp.get_serial_number()

        self.assertEqual(serial_number, expected_serial_number)

    def test_get_base_resolution(self):
        self._library_mock.SH_GetBaseResolution.return_value = 0

        expected_resolution = 0.1
        expected_binsteps = 10

        resolution_patcher = patch('ctypes.c_double', return_value=ctypes.c_double(expected_resolution))
        binsteps_patcher = patch('ctypes.c_int', return_value=ctypes.c_int(expected_binsteps))

        with resolution_patcher, binsteps_patcher:
            resolution, binsteps = self._someharp.get_base_resolution()

        self.assertEqual(resolution, expected_resolution)
        self.assertEqual(binsteps, expected_binsteps)

    def test_get_number_of_input_channels(self):
        self._library_mock.SH_GetNumOfInputChannels.return_value = 0

        expected_channels = 4

        with patch('ctypes.c_int', return_value=ctypes.c_int(expected_channels)):
            channels = self._someharp.get_number_of_input_channels()

        self.assertEqual(channels, expected_channels)

    def test_set_sync_divider(self):

        self._library_mock.SH_SetSyncDiv.return_value = 0
        self._someharp.set_sync_divider(5)
        self._library_mock.SH_SetSyncDiv.assert_called_once_with(0, 5)

    def test_set_input_channel_offset(self):

        self._library_mock.SH_SetInputChannelOffset.return_value = 0
        self._someharp.set_input_channel_offset(1, 99999)
        self._library_mock.SH_SetInputChannelOffset.assert_called_once_with(0, 1, 99999)

    def test_set_input_channel_enable(self):

        self._library_mock.SH_SetInputChannelEnable.return_value = 0
        self._someharp.set_input_channel_enable(0, True)
        self._library_mock.SH_SetInputChannelEnable.assert_called_once_with(0, 0, 1)

    def test_set_stop_overflow(self):

        self._library_mock.SH_SetStopOverflow.return_value = 0
        self._someharp.set_stop_overflow(True, 50000)
        self._library_mock.SH_SetStopOverflow.assert_called_once_with(0, 1, 50000)

    def test_set_binning(self):

        self._library_mock.SH_SetBinning.return_value = 0
        self._someharp.set_binning(3)
        self._library_mock.SH_SetBinning.assert_called_once_with(0, 3)

    def test_set_offset(self):

        self._library_mock.SH_SetOffset.return_value = 0
        self._someharp.set_offset(100000)
        self._library_mock.SH_SetOffset.assert_called_once_with(0, 100000)

    def test_set_histogram_length(self):

        self._library_mock.SH_SetHistoLen.return_value = 0

        expected_actual_length_code = 5

        with patch('ctypes.c_int', return_value=ctypes.c_int(5)) as patcher:
            actual_length = self._someharp.set_histogram_length(6)
            self._library_mock.SH_SetHistoLen.assert_called_once_with(0, 6, patcher.return_value)

        self.assertEqual(actual_length, expected_actual_length_code)

    def test_set_measurement_control(self):
        """Should raise NotImplementedError."""
        with self.assertRaises(NotImplementedError):
            self._someharp.set_measurement_control("a", "b", "c")

    def test_clear_histogram_memory(self):

        self._library_mock.SH_ClearHistMem.return_value = 0
        self._someharp.clear_histogram_memory()
        self._library_mock.SH_ClearHistMem.assert_called_once_with(0)

    def test_start_measurement(self):

        self._library_mock.SH_StartMeas.return_value = 0
        self._someharp.start_measurement(100)
        self._library_mock.SH_StartMeas.assert_called_once_with(0, 100)

    def test_stop_measurement(self):

        self._library_mock.SH_StopMeas.return_value = 0
        self._someharp.stop_measurement()
        self._library_mock.SH_StopMeas.assert_called_once_with(0)

    def test_get_measurement_active(self):

        self._library_mock.SH_CTCStatus.return_value = 0

        with patch('ctypes.c_int', return_value=ctypes.c_int(0)):
            is_running = self._someharp.get_measurement_active()

        self.assertTrue(is_running)

    def test_get_measurement_start_time(self):

        self._library_mock.SH_StartMeas.return_value = 0
        current_time = time.time()

        self._someharp.start_measurement(100)
        posix_time = self._someharp.get_measurement_start_time()
        self.assertAlmostEqual(current_time, posix_time, places=1)

    def test_get_measurement_running_already(self):

        self._library_mock.SH_StartMeas.return_value = 0
        self._someharp.start_measurement(100)
        with self.assertRaises(QMI_InvalidOperationException):
            self._someharp.start_measurement(1)

    def test_get_events(self):

        events = self._someharp.get_events()
        self.assertEqual(len(events), 0)

    def test_get_timestamped_events(self):

        timestamps, events = self._someharp.get_timestamped_events()
        self.assertEqual(timestamps, 0.0)
        self.assertEqual(len(events), 0)

    def test_get_resolution(self):

        self._library_mock.SH_GetResolution.return_value = 0

        expected_resolution = 80.0

        with patch('ctypes.c_double', return_value=ctypes.c_double(expected_resolution)):
            resolution = self._someharp.get_resolution()

        self.assertEqual(resolution, expected_resolution)

    def test_get_sync_rate(self):

        self._library_mock.SH_GetSyncRate.return_value = 0

        expected_sync_rate = 5000

        with patch('ctypes.c_int', return_value=ctypes.c_int(expected_sync_rate)):
            sync_rate = self._someharp.get_sync_rate()

        self.assertEqual(sync_rate, expected_sync_rate)

    def test_get_count_rate(self):

        self._library_mock.SH_GetCountRate.return_value = 0

        expected_count_rate = 5000

        with patch('ctypes.c_int', return_value=ctypes.c_int(expected_count_rate)) as patcher:
            count_rate = self._someharp.get_count_rate(1)
            self._library_mock.SH_GetCountRate.assert_called_once_with(0, 1, patcher.return_value)

        self.assertEqual(count_rate, expected_count_rate)

    def test_get_elapsed_measurement_time(self):

        self._library_mock.SH_GetElapsedMeasTime.return_value = 0

        expected_elapsed = 1000.0

        with patch('ctypes.c_double', return_value=ctypes.c_double(expected_elapsed)):
            elapsed = self._someharp.get_elapsed_measurement_time()

        self.assertEqual(elapsed, expected_elapsed)

    def test_get_sync_period(self):

        self._library_mock.SH_GetSyncPeriod.return_value = 0

        expected_sync_period = 3.14

        with patch('ctypes.c_double', return_value=ctypes.c_double(expected_sync_period)):
            sync_period = self._someharp.get_sync_period()

        self.assertEqual(sync_period, expected_sync_period)

    def test_set_marker_edges(self):

        self._library_mock.SH_SetMarkerEdges.return_value = 0
        self._someharp.set_marker_edges('RISING', 'FALLING', 'RISING', 'FALLING')
        self._library_mock.SH_SetMarkerEdges.assert_called_once_with(0, 1, 0, 1, 0)

    def test_set_marker_enable(self):

        self._library_mock.SH_SetMarkerEnable.return_value = 0
        self._someharp.set_marker_enable(True, False, True, False)
        self._library_mock.SH_SetMarkerEnable.assert_called_once_with(0, 1, 0, 1, 0)

        # ensure bool`s are converted to int`s.
        expect_int_arg = self._library_mock.SH_SetMarkerEnable.call_args_list[0][0][1]
        self.assertEqual(type(expect_int_arg), int)

    def test_set_marker_holdoff_time(self):
        """Test set_marker_holdoff_time function"""
        self._library_mock.SH_SetMarkerHoldoffTime.return_value = 0

        holdoff_time = 25500

        self._someharp.set_marker_holdoff_time(holdoff_time)
        self._library_mock.SH_SetMarkerHoldoffTime.assert_called_once_with(0, holdoff_time)

    def test_get_histogram_with_clear(self):
        """Test on get_histogram function without 'clear' input variable"""
        self._library_mock.SH_GetHistogram.return_value = 0

        self._someharp.get_histogram(1)
        self._library_mock.SH_GetHistogram.assert_called()
        self.assertEqual(len(self._library_mock.SH_GetHistogram.call_args_list[0][0]), 3)

    def get_warnings_text(self):
        """Test get warnings texts can be called and returns text."""
        expected_text = "This is the final warning."
        string_buffer = ctypes.create_string_buffer(expected_text.encode('ASCII'))

        with patch('ctypes.create_string_buffer', return_value=string_buffer):
            warnings_text = self._someharp.get_warnings_text()

        self._someharp.get_warnings_text.assertCalled()
        self.assertEqual(expected_text, warnings_text)

    def test_get_module_info(self):
        """Should raise NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self._someharp.get_module_info()

    def test_set_sync_cfd(self):
        """Should raise NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self._someharp.set_sync_cfd(1, 2)

    def test_set_input_cfd(self):
        """Should raise NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self._someharp.set_input_cfd(1, 2, 3)

    def test_set_sync_offset(self):
        """Should raise NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self._someharp.set_sync_offset(4)

    def test_set_sync_channel_offset(self):
        """Should raise NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self._someharp.set_sync_channel_offset(5)

    def test_get_flags(self):
        """Should raise NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self._someharp.get_flags()

    def get_all_histograms(self):
        """Should raise NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self._someharp.get_all_histograms()

    def get_warnings(self):
        """Should raise NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self._someharp.get_warnings()

    def test_set_trigger_output(self):
        """Should raise NotImplementedError"""
        with self.assertRaises(NotImplementedError):
            self._someharp.set_trigger_output(111)

    def test_set_block_events(self):

        self._someharp.set_block_events(True)

    def test_set_event_filter(self):

        self._someharp.set_event_filter(True)

    def test_set_realtime_histogram(self):

        self._someharp.set_realtime_histogram([0, 1, 2], 1, 100, 0)

    def test_set_realtime_histogram_raises_ValueError(self):

        with self.assertRaises(ValueError):
            self._someharp.set_realtime_histogram([0, 1, 2], 0, 100, 0)

    def test_set_realtime_countrate(self):

        self._someharp.set_realtime_countrate((10, 100), 10)


if __name__ == '__main__':
    unittest.main()



"""Module for a Thorlabs MPC320 motorised fibre polarisation controller."""

from dataclasses import dataclass
import logging
from typing import Dict, List
from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport
from qmi.instruments.thorlabs.apt_packets import (
    HW_GET_INFO,
    MOD_GET_CHANENABLESTATE,
    MOT_GET_USTATUSUPDATE,
    MOT_MOVE_ABSOLUTE,
    MOT_MOVE_COMPLETED,
    MOT_MOVE_HOMED,
    MOT_SET_EEPROMPARAMS,
    POL_GET_SET_PARAMS,
)
from qmi.instruments.thorlabs.apt_protocol import (
    AptChannelJogDirection,
    AptChannelState,
    AptMessageId,
    AptProtocol,
)

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


@dataclass
class Thorlabs_Mpc320_Status:
    """
    Data class for the status of the MPC320

    Attributes:
        channel:        Channel number.
        position:       Absolute position of the channel in degrees.
        velocity:       Velocity in controller units.
        motor_current:  Current of motor in mA
    """

    channel: int
    position: float
    velocity: int
    motor_current: int


@dataclass
class Thorlabs_Mpc320_PolarisationParameters:
    """
    Data class for the polarisation parameters of the MPC320

    Attributes:
        velocity:       The velocity in percentage of the max velocity 400 degrees/s.
        home_position:  The home position of all the paddles/channels in degrees.
        jog_step1:      The position to move paddel/channel 1 by for a jog step in degrees.
        jog_step2:      The position to move paddel/channel 2 by for a jog step in degrees.
        jog_step3:      The position to move paddel/channel 3 by for a jog step in degrees.
    """

    velocity: float
    home_position: float
    jog_step1: float
    jog_step2: float
    jog_step3: float


Thorlabs_Mpc320_ChannelMap: Dict[int, int] = {1: 0x01, 2: 0x02, 3: 0x04}


class Thorlabs_Mpc320(QMI_Instrument):
    """
    Driver for a Thorlabs MPC320 motorised fibre polarisation controller.
    """

    DEFAULT_RESPONSE_TIMEOUT = 1.0

    # the maximum range for a paddle is 170 degrees
    # the value returned by the encoder is 1370 for 170 degrees
    ENCODER_CONVERSION_UNIT = 170 / 1370

    MIN_POSITION_DEGREES = 0
    MAX_POSITION_DEGREES = 170

    MIN_VELOCITY_PERC = 10
    MAX_VELOCITY_PERC = 100

    MIN_CHANNEL_NUMBER = 1
    MAX_CHANNEL_NUMBER = 3

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize the instrument driver.

        Parameters:
            name:       Name for this instrument instance.
            transport:  QMI transport descriptor to connect to the instrument.
        """
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"baudrate": 115200, "rtscts": True})
        self._apt_protocol = AptProtocol(self._transport, default_timeout=self.DEFAULT_RESPONSE_TIMEOUT)

    def _validate_position(self, pos: float) -> None:
        """
        Validate the position. Any position for the MPC320 needs to be in the range 0 to 170 degrees,
        or 0 to 1370 in encoder counts.

        Parameters:
            pos:    Position to validate in degrees.

        Raises:
            QMI_InstrumentException: If the position is invalid.
        """
        if not self.MIN_POSITION_DEGREES <= pos <= self.MAX_POSITION_DEGREES:
            raise QMI_InstrumentException(
                f"Given position {pos} is outside the valid range \
                    [{self.MIN_POSITION_DEGREES}, {self.MAX_POSITION_DEGREES}]"
            )

    def _validate_velocity(self, vel: float) -> None:
        """
        Validate the velocity. Any velocity for the MPC320 needs to be in the range 40 to 400 degrees/s,
        or 10 to 100% of 400 degrees/s.

        Parameters:
            vel:    Velocity to validate in percentage.

        Raises:
            QMI_InstrumentException: if the velocity is invalid.
        """
        if not self.MIN_VELOCITY_PERC <= vel <= self.MAX_VELOCITY_PERC:
            raise QMI_InstrumentException(
                f"Given relative velocity {vel} is outside the valid range \
                    [{self.MIN_VELOCITY_PERC}%, {self.MAX_VELOCITY_PERC}%]"
            )

    def _validate_channel(self, channel_number: int) -> None:
        """
        Validate the channel number. The MPC320 has 3 channels.

        Parameters:
            channel_number: Channel number to validate.

        Raises:
            QMI_InstrumentException: if the channel is not 1, 2 or 3
        """

        if channel_number not in range(self.MIN_CHANNEL_NUMBER, self.MAX_CHANNEL_NUMBER + 1):
            raise QMI_InstrumentException(
                f"Given channel {channel_number} is not in the valid range \
                    [{self.MIN_CHANNEL_NUMBER}, {self.MAX_CHANNEL_NUMBER}]"
            )

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """
        Read instrument type and version and return QMI_InstrumentIdentification instance.

        Returns:
            an instance of QMI_InstrumentIdentification. The version refers to the firmware version.
        """
        _logger.info("[%s] Getting identification of instrument", self._name)
        self._check_is_open()
        # Send request message.
        self._apt_protocol.write_param_command(AptMessageId.HW_REQ_INFO.value)
        # Get response
        resp = self._apt_protocol.ask(HW_GET_INFO)
        return QMI_InstrumentIdentification("Thorlabs", resp.model_number, resp.serial_number, resp.firmware_version)

    @rpc_method
    def identify(self) -> None:
        """
        Identify device by flashing the front panel LEDs.
        """
        _logger.info("[%s] Identifying device", self._name)
        self._check_is_open()
        # Send message.
        # For the MPC320 the channel number does not matter here. The device has one LED that flashes irrespective
        # of the provided channel number.
        self._apt_protocol.write_param_command(AptMessageId.MOD_IDENTIFY.value, 0x01)

    @rpc_method
    def enable_channels(self, channel_numbers: List[int]) -> None:
        """
        Enable the channel(s). Note that this method will disable any channel that is not provided as an argument. For
        example, if you enable channel 1, then 2 and 3 will be disabled. If you have previously enabled a channel(s)
        and fail to include it/them again in this call, that channel(s) will be disabled. For example, if you run the
        following:
        self.enable_channel([1])
        self.enable_channel([2])
        only channel 2 will be enabled and 1 and 3 will be disabled.
        The correct way to call this method in this case is
        self.enable_channel([1,2])

        Parameters:
            channel_number: The channel(s) to enable.
        """
        _logger.info("[%s] Enabling channel(s) %s", self._name, str(channel_numbers))
        self._check_is_open()
        for channel_number in channel_numbers:
            self._validate_channel(channel_number)
        # Make hexadecimal value for channels
        channels_to_enable = 0x00
        for channel_number in channel_numbers:
            channels_to_enable ^= Thorlabs_Mpc320_ChannelMap[channel_number]
        # Send message.
        self._apt_protocol.write_param_command(
            AptMessageId.MOD_SET_CHANENABLESTATE.value,
            channels_to_enable,
            AptChannelState.ENABLE.value,
        )

    @rpc_method
    def disable_all_channels(self) -> None:
        """
        Disable all the channels.
        """
        _logger.info("[%s] Disabling channels", self._name)
        self._check_is_open()
        self._apt_protocol.write_param_command(
            AptMessageId.MOD_SET_CHANENABLESTATE.value,
            0x00,
            AptChannelState.ENABLE.value,
        )

    @rpc_method
    def get_channel_state(self, channel_number: int) -> AptChannelState:
        """
        Get the state of the specified channel.

        Parameters:
            channel_number: The channel to check.

        Returns:
            The state of the channel as an AptChannelState enum.
        """
        _logger.info("[%s] Getting state of channel %d", self._name, channel_number)
        self._validate_channel(channel_number)
        self._check_is_open()
        # Send request message.
        self._apt_protocol.write_param_command(
            AptMessageId.MOD_REQ_CHANENABLESTATE.value,
            Thorlabs_Mpc320_ChannelMap[channel_number],
        )
        # Get response
        resp = self._apt_protocol.ask(MOD_GET_CHANENABLESTATE)
        # For the MPC320 the state 0x01 is the ENABLE state and anything else is DISABLE
        if resp.enable_state == 0x01:
            return AptChannelState.ENABLE
        return AptChannelState.DISABLE

    @rpc_method
    def start_auto_status_update(self) -> None:
        """
        Start automatic status updates from device.
        """
        _logger.info("[%s] Starting automatic status updates from instrument", self._name)
        self._check_is_open()
        # Send message.
        self._apt_protocol.write_param_command(AptMessageId.HW_START_UPDATEMSGS.value)

    @rpc_method
    def stop_auto_status_update(self) -> None:
        """
        Stop automatic status updates from device.
        """
        _logger.info("[%s] Stopping automatic status updates from instrument", self._name)
        self._check_is_open()
        # Send message.
        self._apt_protocol.write_param_command(AptMessageId.HW_STOP_UPDATEMSGS.value)

    @rpc_method
    def home_channel(self, channel_number: int) -> None:
        """
        Start the homing sequence for a given channel.
        After running this command, you must clear the buffer by checking if the channel
        was homed, using is_channel_homed()

        Paramters:
            channel_number: The channel to home.
        """
        _logger.info("[%s] Homing channel %d", self._name, channel_number)
        self._validate_channel(channel_number)
        self._check_is_open()
        # Send message.
        self._apt_protocol.write_param_command(
            AptMessageId.MOT_MOVE_HOME.value, Thorlabs_Mpc320_ChannelMap[channel_number]
        )

    @rpc_method
    def is_channel_homed(self, channel_number: int, timeout: float = DEFAULT_RESPONSE_TIMEOUT) -> bool:
        """
        Check if a given channel is homed. This command should only be run after the method `home_channel`.
        Otherwise you will read bytes from other commands using this method.

        Paramters:
            channel_number: The channel to check.
            timeout:        The time to wait for a response to the homing command
                            with default value DEFAULT_RESPONSE_TIMEOUT.

        Returns:
            True if the channel was homed.
        """
        _logger.info("[%s] Checking if channel %d is homed", self._name, channel_number)
        self._validate_channel(channel_number)
        self._check_is_open()
        # This command needs a workaround. Instead of returning the MOT_MOVE_HOMED message, the instrument returns
        # its current state first, so read that to discard the buffer
        _ = self._apt_protocol.ask(MOT_GET_USTATUSUPDATE, timeout)
        # then read the actual response we need
        try:
            resp = self._apt_protocol.ask(MOT_MOVE_HOMED, timeout)
            return resp.chan_ident == Thorlabs_Mpc320_ChannelMap[channel_number]
        except QMI_TimeoutException:
            _logger.debug("[%s] Channel %d not homed yet", self._name, channel_number)

        return False

    @rpc_method
    def move_absolute(self, channel_number: int, position: float) -> None:
        """
        Move a channel to the specified position. The specified position is in degrees. A conversion is done to convert
        this into encoder counts. This means that there may be a slight mismatch in the specified position and the
        actual position. You may use the get_status_update method to get the actual position.
        After running this command, you must clear the buffer by checking if the channel
        move was completed, using is_move_completed()

        Parameters:
            channel_number: The channel to address.
            position:       Absolute position to move to in degrees.
        """
        _logger.info("[%s] Moving channel %d", self._name, channel_number)
        self._validate_channel(channel_number)
        self._validate_position(position)
        self._check_is_open()
        # Convert position in degrees to encoder counts.
        encoder_position = round(position / self.ENCODER_CONVERSION_UNIT)
        # Make data packet.
        data_packet = MOT_MOVE_ABSOLUTE(
            chan_ident=Thorlabs_Mpc320_ChannelMap[channel_number],
            absolute_distance=encoder_position,
        )
        # Send message.
        self._apt_protocol.write_data_command(AptMessageId.MOT_MOVE_ABSOLUTE.value, data_packet)

    @rpc_method
    def is_move_completed(self, channel_number: int, timeout: float = DEFAULT_RESPONSE_TIMEOUT) -> bool:
        """
        Check if a given channel has completed its move. This command should only be run after a relative or absolute
        move command. Otherwise you will read bytes from other commands.

        Paramters:
            channel_number: The channel to check.
            timeout:        The time to wait for a response to the homing command. This is optional
                            and is set to a default value of DEFAULT_RESPONSE_TIMEOUT.

        Returns:
            True if the move for the channel was completed.
        """
        _logger.info(
            "[%s] Checking if channel %d has completed its move",
            self._name,
            channel_number,
        )
        self._validate_channel(channel_number)
        self._check_is_open()
        # This command needs a workaround. Instead of returning the MOT_MOVE_COMPLETED message, the instrument returns
        # its current state first, so read that to discard the buffer
        _ = self._apt_protocol.ask(MOT_GET_USTATUSUPDATE, timeout)
        # then read the actual response we need
        # then read the actual response we need. If the call times out then the channel has not finished its move.
        try:
            resp = self._apt_protocol.ask(MOT_MOVE_COMPLETED, timeout)
            return resp.chan_ident == Thorlabs_Mpc320_ChannelMap[channel_number]
        except QMI_TimeoutException:
            _logger.debug("[%s] Channel %d move not completed yet", self._name, channel_number)
        return False

    @rpc_method
    def save_parameter_settings(self, channel_number: int, message_id: int) -> None:
        """
        Save parameter settings for a specific message id. These parameters could have been edited via the QMI driver
        or the GUI provided by Thorlabs.

        Parameters:
            channel_number: The channel to address.
            message_id:     ID of message whose parameters need to be saved.
                            Must be provided as a hex number e.g. 0x04B6
        """
        _logger.info("[%s] Saving parameters of message %d", self._name, message_id)
        self._check_is_open()
        self._validate_channel(channel_number)
        # Make data packet.
        data_packet = MOT_SET_EEPROMPARAMS(chan_ident=Thorlabs_Mpc320_ChannelMap[channel_number], msg_id=message_id)
        # Send message.
        self._apt_protocol.write_data_command(AptMessageId.MOT_SET_EEPROMPARAMS.value, data_packet)

    @rpc_method
    def get_status_update(self, channel_number: int) -> Thorlabs_Mpc320_Status:
        """
        Get the status update for a given channel. This call will return the position, velocity, motor current and
        status of the channel.

        Parameters:
            channel_number: The channel to query.

        Returns:
            An instance of Thorlabs_MPC320_Status.
        """
        _logger.info("[%s] Getting position counter of channel %d", self._name, channel_number)
        self._check_is_open()
        self._validate_channel(channel_number)
        # Send request message.
        self._apt_protocol.write_param_command(
            AptMessageId.MOT_REQ_USTATUSUPDATE.value,
            Thorlabs_Mpc320_ChannelMap[channel_number],
        )
        # Get response
        resp = self._apt_protocol.ask(MOT_GET_USTATUSUPDATE)
        return Thorlabs_Mpc320_Status(
            channel=channel_number,
            position=resp.position * self.ENCODER_CONVERSION_UNIT,
            velocity=resp.velocity,
            motor_current=resp.motor_current,
        )

    @rpc_method
    def jog(
        self,
        channel_number: int,
        direction: AptChannelJogDirection = AptChannelJogDirection.FORWARD,
    ) -> None:
        """
        Move a channel specified by its jog step.

        Parameters:
            channel_number: The channel to jog.
            direction:      The direction to jog. This can either be forward or backward. Default is forward.
        """
        _logger.info("[%s] Getting position counter of channel %d", self._name, channel_number)
        self._check_is_open()
        self._validate_channel(channel_number)
        # Send request message.
        self._apt_protocol.write_param_command(
            AptMessageId.MOT_MOVE_JOG.value,
            Thorlabs_Mpc320_ChannelMap[channel_number],
            direction.value,
        )

    @rpc_method
    def set_polarisation_parameters(
        self,
        velocity: int,
        home_pos: int,
        jog_step1: int,
        jog_step2: int,
        jog_step3: int,
    ) -> None:
        """
        Set the polarisation parameters.

        Parameters:
            velocity:       Velocity in range 10% to 100% of 400 degrees/s.
            home_position:  Home position in degrees.
            jog_step1:      Size of jog step for paddle 1.
            jog_step2:      Size of jog step for paddle 2.
            jog_step3:      Size of jog step for paddle 3.
        """
        _logger.info("[%s] Setting polarisation parameters", self._name)
        self._check_is_open()
        # Validate parameters.
        self._validate_velocity(velocity)
        self._validate_position(home_pos)
        self._validate_position(jog_step1)
        self._validate_position(jog_step2)
        self._validate_position(jog_step3)
        # Make data packet.
        data_packet = POL_GET_SET_PARAMS(
            velocity=velocity,
            home_position=round(home_pos / self.ENCODER_CONVERSION_UNIT),
            jog_step1=round(jog_step1 / self.ENCODER_CONVERSION_UNIT),
            jog_step2=round(jog_step2 / self.ENCODER_CONVERSION_UNIT),
            jog_step3=round(jog_step3 / self.ENCODER_CONVERSION_UNIT),
        )
        # Send message.
        self._apt_protocol.write_data_command(AptMessageId.POL_SET_PARAMS.value, data_packet)

    @rpc_method
    def get_polarisation_parameters(self) -> Thorlabs_Mpc320_PolarisationParameters:
        """
        Get the polarisation parameters.

        Returns:
            An instance of Thorlabs_Mpc320_PolarisationParameters
        """
        _logger.info("[%s] Getting polarisation parameters", self._name)
        self._check_is_open()
        # Send request message.
        self._apt_protocol.write_param_command(AptMessageId.POL_REQ_PARAMS.value)
        # Get response.
        params = self._apt_protocol.ask(POL_GET_SET_PARAMS)
        return Thorlabs_Mpc320_PolarisationParameters(
            velocity=params.velocity,
            home_position=params.home_position * self.ENCODER_CONVERSION_UNIT,
            jog_step1=params.jog_step1 * self.ENCODER_CONVERSION_UNIT,
            jog_step2=params.jog_step2 * self.ENCODER_CONVERSION_UNIT,
            jog_step3=params.jog_step3 * self.ENCODER_CONVERSION_UNIT,
        )



from enum import IntEnum

import numpy as np

SYNC_TYPE = 64  # Event type of SYNC events.
# Numpy record type used to represent event records.
# Event types 0 to 7 represent a normal event.
# Event type 64 represents a SYNC event.
# Event types 65 to 79 represent marker events.
EventDataType = np.dtype([
    ("type", np.uint8),
    ("timestamp", np.uint64)
])


class EventFilterMode(IntEnum):
    """The filter modes for channel filtering.

    0: NO_EVENTS - block all events from a channel.
    1: ALL_EVENTS - Do not filter events from this channel.
    2: APERTURE - filters events outside given 'aperture' time range (min-max) per sync event.
    """
    NO_EVENTS = 0
    ALL_EVENTS = 1
    APERTURE = 2


class EventDecoder:

    def __init__(self, sync_frequency_hz: float = 5E6, resolution_ps: float = 1.0) -> None:
        """EventDecoder base class for different mode event decoders.

        Parameters:
            sync_frequency_hz:     The sync signal frequency is Hz. Default 5MHz.
            resolution_ps:         The time tag resolution from 1 to 33 554 432 ps. Relevant only in T3 mode, as in T2
            mode this is always 1 (base resolution, default).
        """
        self._sync_period_ps = 1E12 / sync_frequency_hz  # From 1/s to 1/ps. Used only in T3 mode
        self._resolution_ps = resolution_ps
        self._overflow_counter = 0

    def process_data(self, fifo_data: np.ndarray) -> np.ndarray:
        """Decode event records from the Multi- or HydraHarp T2 or T3 data stream.

        This function decodes raw 32-bit TTTR event records and returns
        an array of structured event records.

        Overflow records in the TTTR stream are used to assign an unwrapped
        64-bit timestamp value to each event. The overflow records are removed
        from the event array returned by this function.

        The `EventDecoder` instance keeps track of the running overflow counter.
        The sequence of calls to this method must therefore correspond to
        a single TTTR data stream, processed in order, without gaps or data loss.

        Parameters:
            fifo_data: Numpy array of 32-bit raw event records in TTTR T2 or T3 format.

        Returns:
            Numpy array of `EventDataType` records.
        """
        raise NotImplementedError


class _T2EventDecoder(EventDecoder):
    """Decode the MultiHarp, HydraHarp TTTR data stream in T2 mode."""

    def process_data(self, fifo_data: np.ndarray) -> np.ndarray:

        #
        # MultiHarp T2 TTTR event record format:
        #   bit  31    = special_flag
        #   bits 30:25 = channel
        #   bits 24:0  = timetag
        #
        # The "timetag" value is a multiple of the instrument base resolution (80 ps for MultiHarp).
        # Overflow of the 25-bit counter is marked via special overflow records.
        #
        # If special_flag == 0, the record represents a normal event on the specified channel.
        # If special_flag == 1 And channel == 63, the record represents overflow of the 25-bit time value.
        # If special_flag == 1 And channel == 0, the record represents a SYNC event.
        # If special_flag == 1 And 1 <= channel <= 15, the record represents a marker event.
        #

        # Type code for overflow events.
        # Note that overflow channel is 63, but here also special flag is added, so 0x7f.
        overflow_type = 0x7f

        # Overflow period is the full range of the 25 bits time tag.
        overflow_period = (1 << 25)

        # Extract the 7-bit "type" field of each record.
        # The "type" field consists of the 1-bit "special" flag together with the 6-bit "channel" field.
        record_types = (fifo_data >> 25).astype(np.uint8)

        # Extract the 25-bit "timetag" field. For overflow records, this is the overflow increment;
        # for event records, this is the time-tag (offset since last overflow).
        record_tags = (fifo_data & 0x01ffffff)  # 0x01ffffff = 2^25 - 1

        # Make a bool array marking the overflow records.
        overflow_records_bool = (record_types == overflow_type)

        # Make a bool array marking the normal (non-overflow) records.
        event_records_bool = ~overflow_records_bool

        # Prepare a running count of the number of overflows
        overflow_counts = np.cumsum(overflow_records_bool * record_tags, dtype=np.uint64)
        overflow_counts += self._overflow_counter

        # Update the overflow counter to take this batch of events into account.
        if len(overflow_counts) > 0:
            self._overflow_counter = int(overflow_counts[-1])

        # Select the event types of the non-overflow records.
        event_types = record_types[event_records_bool]

        # Calculate the event timestamps of the non-overflow records.
        event_timestamps = overflow_counts[event_records_bool] * overflow_period + record_tags[event_records_bool]

        # Copy the non-overflow records to a result array.
        events = np.empty(len(event_types), dtype=EventDataType)
        events["type"] = event_types  # type: ignore[call-overload]
        events["timestamp"] = event_timestamps  # type: ignore[call-overload]

        return events


class _T3EventDecoder(EventDecoder):
    """Decode the MultiHarp, HydraHarp TTTR data stream in T3 mode."""

    def process_data(self, fifo_data: np.ndarray) -> np.ndarray:

        # Multi/HydraHarp T3 TTTR event record format:
        #   bit  31    = special_flag
        #   bits 30:25 = channel
        #   bits 24:10 = d_time
        #   bits 9:0   = n_sync
        #
        # The "n_sync" and "d_time" values are multiples of the instrument base resolution (32 ps for HydraHarp).
        # Overflow of the 25-bit counter is marked via special overflow records.
        #
        # If special_flag == 0, the record represents a normal event record on the specified channel.
        # If special_flag == 1 And channel == 63, the record represents sync count overflow of the 25-bit time value.
        # The number of overflows can be read from the n_sync value.
        # If special_flag == 1 And 1 <= channel <= 15, the record represents an external marker event.
        #

        # Type code for overflow events.
        # Note that overflow channel is 63, but here also special flag is added, so 0x7f.
        overflow_type = 0x7f

        # Overflow period is the full range of the 10 bits sync counter times the period of the sync signal.
        overflow_period = (1 << 10) * self._sync_period_ps

        # Extract the 7-bit "type" field of each record by shifting 25 last bits out.
        # The "type" field consists of the 1-bit "special" flag together with the 6-bit "channel" field.
        record_types = (fifo_data >> 25).astype(np.uint8)

        # Extract the 15-bit "d_time" field by shifting 10 last bits out and taking 15 bits. Include resolution
        d_times = (fifo_data >> 10 & 0x07fff).astype(np.uint64) * self._resolution_ps  # 0x07fff = 2^15 - 1
        # Extract the 10-bit "n_sync" field. For overflow records, this is the overflow increment;
        # for event records, this is the time-tag (non-frequency corrected sync moment).
        n_syncs = (fifo_data & 0x03ff)  # 0x03ff = 2^10 - 1

        # Make a bool array marking the overflow records.
        overflow_records_bool = (record_types == overflow_type)

        # Make a bool array marking the normal (non-overflow) records.
        event_records_bool = ~overflow_records_bool

        # Prepare a running count of the number of overflows
        overflow_counts = np.cumsum(overflow_records_bool * n_syncs, dtype=np.uint64)
        overflow_counts += self._overflow_counter

        # Update the overflow counter to take this batch of events into account.
        if len(overflow_counts) > 0:
            self._overflow_counter = int(overflow_counts[-1])

        # Select the event types of the non-overflow records.
        event_types = record_types[event_records_bool]

        # Calculate the sync and event timestamps of the non-overflow records.
        sync_timestamps = overflow_counts[event_records_bool] * overflow_period +\
                          n_syncs[event_records_bool] * self._sync_period_ps
        event_timestamps = sync_timestamps + d_times[event_records_bool]

        # Remove duplicates from syncs
        sync_timestamps = np.unique(sync_timestamps)

        # Copy the non-overflow records to a result array.
        num_data_events = len(event_types)
        num_events = num_data_events + len(sync_timestamps)
        events = np.empty(num_events, dtype=EventDataType)
        events[:num_data_events]["type"] = event_types  # type: ignore[call-overload]
        events[:num_data_events]["timestamp"] = event_timestamps  # type: ignore[call-overload]
        events[num_data_events:]["type"] = 64  # type: ignore[call-overload]
        events[num_data_events:]["timestamp"] = sync_timestamps  # type: ignore[call-overload]
        # Events need to be sorted by timestamp, and by type number so that sync# 64 is before channel#.
        # NOTE: Since numpy 2.0 overflow of values is not allowed anymore. Therefore, for the means of the lexical sort
        # where we use the trick of setting the event types as their negatives to get the right order, now must be
        # retyped from unsigned 8-bit integer to signed 16-bit integer. The returned array by the lexsort does
        # re-conform to uint8 when placing it back to `events`, so no errors should be introduced.
        events = events[np.lexsort((-1 * events["type"].astype("int16"), events["timestamp"]))]

        return events
